#ifndef random_systems_heptic_symmetric_system_hpp
#define random_systems_heptic_symmetric_system_hpp

#include <utility>
#include <iterator>
#include <math.h>

namespace flopoco
{
namespace random
{
namespace systems
 {

template<class T,class TACC=T>
class SymmetricHepticSystem{
private:
  T ss[57];
  T tmom[9];
  template<class TT> TT square(const TT&x) const { return x*x; }
public:
  typedef T real_t;

  int arity() const { return 4; };

  SymmetricHepticSystem(  const T *_ss ,   const T *_tmom )
  {
    std::copy(_ss, _ss+57, ss);
    std::copy(_tmom, _tmom+9, tmom);
  }

  void DefaultGuess(T *params) const{
    params[0]=1; // c1
    params[1]=0; // c3
    params[2]=0; // c5
    params[3]=0; // c7
  }

T metric(const T *parameters) const {
  T c1p1 = parameters[0];
  T c3p1 = parameters[1];
  T c5p1 = parameters[2];
  T c7p1 = parameters[3];
  T c1p2 = c1p1*c1p1;
  T c3p2 = c3p1*c3p1;
  T c5p2 = c5p1*c5p1;
  T c7p2 = c7p1*c7p1;
  TACC root_0 = convert<TACC>(2*c1p1*c3p1*ss[4]);
  root_0 += 2*c1p1*c5p1*ss[6];
  root_0 += 2*c1p1*c7p1*ss[8];
  root_0 += 2*c3p1*c5p1*ss[8];
  root_0 += 2*c3p1*c7p1*ss[10];
  root_0 += 2*c5p1*c7p1*ss[12];
  root_0 += c1p2*ss[2];
  root_0 += c3p2*ss[6];
  root_0 += c5p2*ss[10];
  root_0 += c7p2*ss[14];
  root_0 += -tmom[2];
  T c1p3 = c1p2*c1p1;
  T c1p4 = c1p3*c1p1;
  T c3p3 = c3p2*c3p1;
  T c3p4 = c3p3*c3p1;
  T c5p3 = c5p2*c5p1;
  T c5p4 = c5p3*c5p1;
  T c7p3 = c7p2*c7p1;
  T c7p4 = c7p3*c7p1;
  TACC root_1 = convert<TACC>(24*c1p1*c3p1*c5p1*c7p1*ss[16]);
  root_1 += 12*c1p1*c3p1*c5p2*ss[14];
  root_1 += 12*c1p1*c3p1*c7p2*ss[18];
  root_1 += 12*c1p1*c3p2*c5p1*ss[12];
  root_1 += 12*c1p1*c3p2*c7p1*ss[14];
  root_1 += 12*c1p1*c5p1*c7p2*ss[20];
  root_1 += 12*c1p1*c5p2*c7p1*ss[18];
  root_1 += 12*c1p2*c3p1*c5p1*ss[10];
  root_1 += 12*c1p2*c3p1*c7p1*ss[12];
  root_1 += 12*c1p2*c5p1*c7p1*ss[14];
  root_1 += 12*c3p1*c5p1*c7p2*ss[22];
  root_1 += 12*c3p1*c5p2*c7p1*ss[20];
  root_1 += 12*c3p2*c5p1*c7p1*ss[18];
  root_1 += 4*c1p1*c3p3*ss[10];
  root_1 += 4*c1p1*c5p3*ss[16];
  root_1 += 4*c1p1*c7p3*ss[22];
  root_1 += 6*c1p2*c3p2*ss[8];
  root_1 += 6*c1p2*c5p2*ss[12];
  root_1 += 6*c1p2*c7p2*ss[16];
  root_1 += 4*c1p3*c3p1*ss[6];
  root_1 += 4*c1p3*c5p1*ss[8];
  root_1 += 4*c1p3*c7p1*ss[10];
  root_1 += 4*c3p1*c5p3*ss[18];
  root_1 += 4*c3p1*c7p3*ss[24];
  root_1 += 6*c3p2*c5p2*ss[16];
  root_1 += 6*c3p2*c7p2*ss[20];
  root_1 += 4*c3p3*c5p1*ss[14];
  root_1 += 4*c3p3*c7p1*ss[16];
  root_1 += 4*c5p1*c7p3*ss[26];
  root_1 += 6*c5p2*c7p2*ss[24];
  root_1 += 4*c5p3*c7p1*ss[22];
  root_1 += c1p4*ss[4];
  root_1 += c3p4*ss[12];
  root_1 += c5p4*ss[20];
  root_1 += c7p4*ss[28];
  root_1 += -tmom[4];
  T c1p5 = c1p4*c1p1;
  T c1p6 = c1p5*c1p1;
  T c3p5 = c3p4*c3p1;
  T c3p6 = c3p5*c3p1;
  T c5p5 = c5p4*c5p1;
  T c5p6 = c5p5*c5p1;
  T c7p5 = c7p4*c7p1;
  T c7p6 = c7p5*c7p1;
  TACC root_2 = convert<TACC>(120*c1p1*c3p1*c5p1*c7p3*ss[30]);
  root_2 += 180*c1p1*c3p1*c5p2*c7p2*ss[28];
  root_2 += 120*c1p1*c3p1*c5p3*c7p1*ss[26];
  root_2 += 180*c1p1*c3p2*c5p1*c7p2*ss[26];
  root_2 += 180*c1p1*c3p2*c5p2*c7p1*ss[24];
  root_2 += 120*c1p1*c3p3*c5p1*c7p1*ss[22];
  root_2 += 180*c1p2*c3p1*c5p1*c7p2*ss[24];
  root_2 += 180*c1p2*c3p1*c5p2*c7p1*ss[22];
  root_2 += 180*c1p2*c3p2*c5p1*c7p1*ss[20];
  root_2 += 120*c1p3*c3p1*c5p1*c7p1*ss[18];
  root_2 += 30*c1p1*c3p1*c5p4*ss[24];
  root_2 += 30*c1p1*c3p1*c7p4*ss[32];
  root_2 += 60*c1p1*c3p2*c5p3*ss[22];
  root_2 += 60*c1p1*c3p2*c7p3*ss[28];
  root_2 += 60*c1p1*c3p3*c5p2*ss[20];
  root_2 += 60*c1p1*c3p3*c7p2*ss[24];
  root_2 += 30*c1p1*c3p4*c5p1*ss[18];
  root_2 += 30*c1p1*c3p4*c7p1*ss[20];
  root_2 += 30*c1p1*c5p1*c7p4*ss[34];
  root_2 += 60*c1p1*c5p2*c7p3*ss[32];
  root_2 += 60*c1p1*c5p3*c7p2*ss[30];
  root_2 += 30*c1p1*c5p4*c7p1*ss[28];
  root_2 += 60*c1p2*c3p1*c5p3*ss[20];
  root_2 += 60*c1p2*c3p1*c7p3*ss[26];
  root_2 += 90*c1p2*c3p2*c5p2*ss[18];
  root_2 += 90*c1p2*c3p2*c7p2*ss[22];
  root_2 += 60*c1p2*c3p3*c5p1*ss[16];
  root_2 += 60*c1p2*c3p3*c7p1*ss[18];
  root_2 += 60*c1p2*c5p1*c7p3*ss[28];
  root_2 += 90*c1p2*c5p2*c7p2*ss[26];
  root_2 += 60*c1p2*c5p3*c7p1*ss[24];
  root_2 += 60*c1p3*c3p1*c5p2*ss[16];
  root_2 += 60*c1p3*c3p1*c7p2*ss[20];
  root_2 += 60*c1p3*c3p2*c5p1*ss[14];
  root_2 += 60*c1p3*c3p2*c7p1*ss[16];
  root_2 += 60*c1p3*c5p1*c7p2*ss[22];
  root_2 += 60*c1p3*c5p2*c7p1*ss[20];
  root_2 += 30*c1p4*c3p1*c5p1*ss[12];
  root_2 += 30*c1p4*c3p1*c7p1*ss[14];
  root_2 += 30*c1p4*c5p1*c7p1*ss[16];
  root_2 += 30*c3p1*c5p1*c7p4*ss[36];
  root_2 += 60*c3p1*c5p2*c7p3*ss[34];
  root_2 += 60*c3p1*c5p3*c7p2*ss[32];
  root_2 += 30*c3p1*c5p4*c7p1*ss[30];
  root_2 += 60*c3p2*c5p1*c7p3*ss[32];
  root_2 += 90*c3p2*c5p2*c7p2*ss[30];
  root_2 += 60*c3p2*c5p3*c7p1*ss[28];
  root_2 += 60*c3p3*c5p1*c7p2*ss[28];
  root_2 += 60*c3p3*c5p2*c7p1*ss[26];
  root_2 += 30*c3p4*c5p1*c7p1*ss[24];
  root_2 += 6*c1p1*c3p5*ss[16];
  root_2 += 6*c1p1*c5p5*ss[26];
  root_2 += 6*c1p1*c7p5*ss[36];
  root_2 += 15*c1p2*c3p4*ss[14];
  root_2 += 15*c1p2*c5p4*ss[22];
  root_2 += 15*c1p2*c7p4*ss[30];
  root_2 += 20*c1p3*c3p3*ss[12];
  root_2 += 20*c1p3*c5p3*ss[18];
  root_2 += 20*c1p3*c7p3*ss[24];
  root_2 += 15*c1p4*c3p2*ss[10];
  root_2 += 15*c1p4*c5p2*ss[14];
  root_2 += 15*c1p4*c7p2*ss[18];
  root_2 += 6*c1p5*c3p1*ss[8];
  root_2 += 6*c1p5*c5p1*ss[10];
  root_2 += 6*c1p5*c7p1*ss[12];
  root_2 += 6*c3p1*c5p5*ss[28];
  root_2 += 6*c3p1*c7p5*ss[38];
  root_2 += 15*c3p2*c5p4*ss[26];
  root_2 += 15*c3p2*c7p4*ss[34];
  root_2 += 20*c3p3*c5p3*ss[24];
  root_2 += 20*c3p3*c7p3*ss[30];
  root_2 += 15*c3p4*c5p2*ss[22];
  root_2 += 15*c3p4*c7p2*ss[26];
  root_2 += 6*c3p5*c5p1*ss[20];
  root_2 += 6*c3p5*c7p1*ss[22];
  root_2 += 6*c5p1*c7p5*ss[40];
  root_2 += 15*c5p2*c7p4*ss[38];
  root_2 += 20*c5p3*c7p3*ss[36];
  root_2 += 15*c5p4*c7p2*ss[34];
  root_2 += 6*c5p5*c7p1*ss[32];
  root_2 += c1p6*ss[6];
  root_2 += c3p6*ss[18];
  root_2 += c5p6*ss[30];
  root_2 += c7p6*ss[42];
  root_2 += -tmom[6];
  T c1p7 = c1p6*c1p1;
  T c1p8 = c1p7*c1p1;
  T c3p7 = c3p6*c3p1;
  T c3p8 = c3p7*c3p1;
  T c5p7 = c5p6*c5p1;
  T c5p8 = c5p7*c5p1;
  T c7p7 = c7p6*c7p1;
  T c7p8 = c7p7*c7p1;
  TACC root_3 = convert<TACC>(336*c1p1*c3p1*c5p1*c7p5*ss[44]);
  root_3 += 840*c1p1*c3p1*c5p2*c7p4*ss[42];
  root_3 += 1120*c1p1*c3p1*c5p3*c7p3*ss[40];
  root_3 += 840*c1p1*c3p1*c5p4*c7p2*ss[38];
  root_3 += 336*c1p1*c3p1*c5p5*c7p1*ss[36];
  root_3 += 840*c1p1*c3p2*c5p1*c7p4*ss[40];
  root_3 += 1680*c1p1*c3p2*c5p2*c7p3*ss[38];
  root_3 += 1680*c1p1*c3p2*c5p3*c7p2*ss[36];
  root_3 += 840*c1p1*c3p2*c5p4*c7p1*ss[34];
  root_3 += 1120*c1p1*c3p3*c5p1*c7p3*ss[36];
  root_3 += 1680*c1p1*c3p3*c5p2*c7p2*ss[34];
  root_3 += 1120*c1p1*c3p3*c5p3*c7p1*ss[32];
  root_3 += 840*c1p1*c3p4*c5p1*c7p2*ss[32];
  root_3 += 840*c1p1*c3p4*c5p2*c7p1*ss[30];
  root_3 += 336*c1p1*c3p5*c5p1*c7p1*ss[28];
  root_3 += 840*c1p2*c3p1*c5p1*c7p4*ss[38];
  root_3 += 1680*c1p2*c3p1*c5p2*c7p3*ss[36];
  root_3 += 1680*c1p2*c3p1*c5p3*c7p2*ss[34];
  root_3 += 840*c1p2*c3p1*c5p4*c7p1*ss[32];
  root_3 += 1680*c1p2*c3p2*c5p1*c7p3*ss[34];
  root_3 += 2520*c1p2*c3p2*c5p2*c7p2*ss[32];
  root_3 += 1680*c1p2*c3p2*c5p3*c7p1*ss[30];
  root_3 += 1680*c1p2*c3p3*c5p1*c7p2*ss[30];
  root_3 += 1680*c1p2*c3p3*c5p2*c7p1*ss[28];
  root_3 += 840*c1p2*c3p4*c5p1*c7p1*ss[26];
  root_3 += 1120*c1p3*c3p1*c5p1*c7p3*ss[32];
  root_3 += 1680*c1p3*c3p1*c5p2*c7p2*ss[30];
  root_3 += 1120*c1p3*c3p1*c5p3*c7p1*ss[28];
  root_3 += 1680*c1p3*c3p2*c5p1*c7p2*ss[28];
  root_3 += 1680*c1p3*c3p2*c5p2*c7p1*ss[26];
  root_3 += 1120*c1p3*c3p3*c5p1*c7p1*ss[24];
  root_3 += 840*c1p4*c3p1*c5p1*c7p2*ss[26];
  root_3 += 840*c1p4*c3p1*c5p2*c7p1*ss[24];
  root_3 += 840*c1p4*c3p2*c5p1*c7p1*ss[22];
  root_3 += 336*c1p5*c3p1*c5p1*c7p1*ss[20];
  root_3 += 56*c1p1*c3p1*c5p6*ss[34];
  root_3 += 56*c1p1*c3p1*c7p6*ss[46];
  root_3 += 168*c1p1*c3p2*c5p5*ss[32];
  root_3 += 168*c1p1*c3p2*c7p5*ss[42];
  root_3 += 280*c1p1*c3p3*c5p4*ss[30];
  root_3 += 280*c1p1*c3p3*c7p4*ss[38];
  root_3 += 280*c1p1*c3p4*c5p3*ss[28];
  root_3 += 280*c1p1*c3p4*c7p3*ss[34];
  root_3 += 168*c1p1*c3p5*c5p2*ss[26];
  root_3 += 168*c1p1*c3p5*c7p2*ss[30];
  root_3 += 56*c1p1*c3p6*c5p1*ss[24];
  root_3 += 56*c1p1*c3p6*c7p1*ss[26];
  root_3 += 56*c1p1*c5p1*c7p6*ss[48];
  root_3 += 168*c1p1*c5p2*c7p5*ss[46];
  root_3 += 280*c1p1*c5p3*c7p4*ss[44];
  root_3 += 280*c1p1*c5p4*c7p3*ss[42];
  root_3 += 168*c1p1*c5p5*c7p2*ss[40];
  root_3 += 56*c1p1*c5p6*c7p1*ss[38];
  root_3 += 168*c1p2*c3p1*c5p5*ss[30];
  root_3 += 168*c1p2*c3p1*c7p5*ss[40];
  root_3 += 420*c1p2*c3p2*c5p4*ss[28];
  root_3 += 420*c1p2*c3p2*c7p4*ss[36];
  root_3 += 560*c1p2*c3p3*c5p3*ss[26];
  root_3 += 560*c1p2*c3p3*c7p3*ss[32];
  root_3 += 420*c1p2*c3p4*c5p2*ss[24];
  root_3 += 420*c1p2*c3p4*c7p2*ss[28];
  root_3 += 168*c1p2*c3p5*c5p1*ss[22];
  root_3 += 168*c1p2*c3p5*c7p1*ss[24];
  root_3 += 168*c1p2*c5p1*c7p5*ss[42];
  root_3 += 420*c1p2*c5p2*c7p4*ss[40];
  root_3 += 560*c1p2*c5p3*c7p3*ss[38];
  root_3 += 420*c1p2*c5p4*c7p2*ss[36];
  root_3 += 168*c1p2*c5p5*c7p1*ss[34];
  root_3 += 280*c1p3*c3p1*c5p4*ss[26];
  root_3 += 280*c1p3*c3p1*c7p4*ss[34];
  root_3 += 560*c1p3*c3p2*c5p3*ss[24];
  root_3 += 560*c1p3*c3p2*c7p3*ss[30];
  root_3 += 560*c1p3*c3p3*c5p2*ss[22];
  root_3 += 560*c1p3*c3p3*c7p2*ss[26];
  root_3 += 280*c1p3*c3p4*c5p1*ss[20];
  root_3 += 280*c1p3*c3p4*c7p1*ss[22];
  root_3 += 280*c1p3*c5p1*c7p4*ss[36];
  root_3 += 560*c1p3*c5p2*c7p3*ss[34];
  root_3 += 560*c1p3*c5p3*c7p2*ss[32];
  root_3 += 280*c1p3*c5p4*c7p1*ss[30];
  root_3 += 280*c1p4*c3p1*c5p3*ss[22];
  root_3 += 280*c1p4*c3p1*c7p3*ss[28];
  root_3 += 420*c1p4*c3p2*c5p2*ss[20];
  root_3 += 420*c1p4*c3p2*c7p2*ss[24];
  root_3 += 280*c1p4*c3p3*c5p1*ss[18];
  root_3 += 280*c1p4*c3p3*c7p1*ss[20];
  root_3 += 280*c1p4*c5p1*c7p3*ss[30];
  root_3 += 420*c1p4*c5p2*c7p2*ss[28];
  root_3 += 280*c1p4*c5p3*c7p1*ss[26];
  root_3 += 168*c1p5*c3p1*c5p2*ss[18];
  root_3 += 168*c1p5*c3p1*c7p2*ss[22];
  root_3 += 168*c1p5*c3p2*c5p1*ss[16];
  root_3 += 168*c1p5*c3p2*c7p1*ss[18];
  root_3 += 168*c1p5*c5p1*c7p2*ss[24];
  root_3 += 168*c1p5*c5p2*c7p1*ss[22];
  root_3 += 56*c1p6*c3p1*c5p1*ss[14];
  root_3 += 56*c1p6*c3p1*c7p1*ss[16];
  root_3 += 56*c1p6*c5p1*c7p1*ss[18];
  root_3 += 56*c3p1*c5p1*c7p6*ss[50];
  root_3 += 168*c3p1*c5p2*c7p5*ss[48];
  root_3 += 280*c3p1*c5p3*c7p4*ss[46];
  root_3 += 280*c3p1*c5p4*c7p3*ss[44];
  root_3 += 168*c3p1*c5p5*c7p2*ss[42];
  root_3 += 56*c3p1*c5p6*c7p1*ss[40];
  root_3 += 168*c3p2*c5p1*c7p5*ss[46];
  root_3 += 420*c3p2*c5p2*c7p4*ss[44];
  root_3 += 560*c3p2*c5p3*c7p3*ss[42];
  root_3 += 420*c3p2*c5p4*c7p2*ss[40];
  root_3 += 168*c3p2*c5p5*c7p1*ss[38];
  root_3 += 280*c3p3*c5p1*c7p4*ss[42];
  root_3 += 560*c3p3*c5p2*c7p3*ss[40];
  root_3 += 560*c3p3*c5p3*c7p2*ss[38];
  root_3 += 280*c3p3*c5p4*c7p1*ss[36];
  root_3 += 280*c3p4*c5p1*c7p3*ss[38];
  root_3 += 420*c3p4*c5p2*c7p2*ss[36];
  root_3 += 280*c3p4*c5p3*c7p1*ss[34];
  root_3 += 168*c3p5*c5p1*c7p2*ss[34];
  root_3 += 168*c3p5*c5p2*c7p1*ss[32];
  root_3 += 56*c3p6*c5p1*c7p1*ss[30];
  root_3 += 8*c1p1*c3p7*ss[22];
  root_3 += 8*c1p1*c5p7*ss[36];
  root_3 += 8*c1p1*c7p7*ss[50];
  root_3 += 28*c1p2*c3p6*ss[20];
  root_3 += 28*c1p2*c5p6*ss[32];
  root_3 += 28*c1p2*c7p6*ss[44];
  root_3 += 56*c1p3*c3p5*ss[18];
  root_3 += 56*c1p3*c5p5*ss[28];
  root_3 += 56*c1p3*c7p5*ss[38];
  root_3 += 70*c1p4*c3p4*ss[16];
  root_3 += 70*c1p4*c5p4*ss[24];
  root_3 += 70*c1p4*c7p4*ss[32];
  root_3 += 56*c1p5*c3p3*ss[14];
  root_3 += 56*c1p5*c5p3*ss[20];
  root_3 += 56*c1p5*c7p3*ss[26];
  root_3 += 28*c1p6*c3p2*ss[12];
  root_3 += 28*c1p6*c5p2*ss[16];
  root_3 += 28*c1p6*c7p2*ss[20];
  root_3 += 8*c1p7*c3p1*ss[10];
  root_3 += 8*c1p7*c5p1*ss[12];
  root_3 += 8*c1p7*c7p1*ss[14];
  root_3 += 8*c3p1*c5p7*ss[38];
  root_3 += 8*c3p1*c7p7*ss[52];
  root_3 += 28*c3p2*c5p6*ss[36];
  root_3 += 28*c3p2*c7p6*ss[48];
  root_3 += 56*c3p3*c5p5*ss[34];
  root_3 += 56*c3p3*c7p5*ss[44];
  root_3 += 70*c3p4*c5p4*ss[32];
  root_3 += 70*c3p4*c7p4*ss[40];
  root_3 += 56*c3p5*c5p3*ss[30];
  root_3 += 56*c3p5*c7p3*ss[36];
  root_3 += 28*c3p6*c5p2*ss[28];
  root_3 += 28*c3p6*c7p2*ss[32];
  root_3 += 8*c3p7*c5p1*ss[26];
  root_3 += 8*c3p7*c7p1*ss[28];
  root_3 += 8*c5p1*c7p7*ss[54];
  root_3 += 28*c5p2*c7p6*ss[52];
  root_3 += 56*c5p3*c7p5*ss[50];
  root_3 += 70*c5p4*c7p4*ss[48];
  root_3 += 56*c5p5*c7p3*ss[46];
  root_3 += 28*c5p6*c7p2*ss[44];
  root_3 += 8*c5p7*c7p1*ss[42];
  root_3 += c1p8*ss[8];
  root_3 += c3p8*ss[24];
  root_3 += c5p8*ss[40];
  root_3 += c7p8*ss[56];
  root_3 += -tmom[8];
  return convert<T>(sqrt(root_0*root_0+root_1*root_1+root_2*root_2+root_3*root_3));
}
T metric_and_gradient(const T *parameters, T *gradient) const {
  T c1p1 = parameters[0];
  T c3p1 = parameters[1];
  T c5p1 = parameters[2];
  T c7p1 = parameters[3];
  T c1p2 = c1p1*c1p1;
  T c3p2 = c3p1*c3p1;
  T c5p2 = c5p1*c5p1;
  T c7p2 = c7p1*c7p1;
  TACC sys_part_0 = convert<TACC>(2*c1p1*c3p1*ss[4]);
  sys_part_0 += 2*c1p1*c5p1*ss[6];
  sys_part_0 += 2*c1p1*c7p1*ss[8];
  sys_part_0 += 2*c3p1*c5p1*ss[8];
  sys_part_0 += 2*c3p1*c7p1*ss[10];
  sys_part_0 += 2*c5p1*c7p1*ss[12];
  sys_part_0 += c1p2*ss[2];
  sys_part_0 += c3p2*ss[6];
  sys_part_0 += c5p2*ss[10];
  sys_part_0 += c7p2*ss[14];
  sys_part_0 += -tmom[2];
  T c1p3 = c1p2*c1p1;
  T c1p4 = c1p3*c1p1;
  T c3p3 = c3p2*c3p1;
  T c3p4 = c3p3*c3p1;
  T c5p3 = c5p2*c5p1;
  T c5p4 = c5p3*c5p1;
  T c7p3 = c7p2*c7p1;
  T c7p4 = c7p3*c7p1;
  TACC sys_part_1 = convert<TACC>(24*c1p1*c3p1*c5p1*c7p1*ss[16]);
  sys_part_1 += 12*c1p1*c3p1*c5p2*ss[14];
  sys_part_1 += 12*c1p1*c3p1*c7p2*ss[18];
  sys_part_1 += 12*c1p1*c3p2*c5p1*ss[12];
  sys_part_1 += 12*c1p1*c3p2*c7p1*ss[14];
  sys_part_1 += 12*c1p1*c5p1*c7p2*ss[20];
  sys_part_1 += 12*c1p1*c5p2*c7p1*ss[18];
  sys_part_1 += 12*c1p2*c3p1*c5p1*ss[10];
  sys_part_1 += 12*c1p2*c3p1*c7p1*ss[12];
  sys_part_1 += 12*c1p2*c5p1*c7p1*ss[14];
  sys_part_1 += 12*c3p1*c5p1*c7p2*ss[22];
  sys_part_1 += 12*c3p1*c5p2*c7p1*ss[20];
  sys_part_1 += 12*c3p2*c5p1*c7p1*ss[18];
  sys_part_1 += 4*c1p1*c3p3*ss[10];
  sys_part_1 += 4*c1p1*c5p3*ss[16];
  sys_part_1 += 4*c1p1*c7p3*ss[22];
  sys_part_1 += 6*c1p2*c3p2*ss[8];
  sys_part_1 += 6*c1p2*c5p2*ss[12];
  sys_part_1 += 6*c1p2*c7p2*ss[16];
  sys_part_1 += 4*c1p3*c3p1*ss[6];
  sys_part_1 += 4*c1p3*c5p1*ss[8];
  sys_part_1 += 4*c1p3*c7p1*ss[10];
  sys_part_1 += 4*c3p1*c5p3*ss[18];
  sys_part_1 += 4*c3p1*c7p3*ss[24];
  sys_part_1 += 6*c3p2*c5p2*ss[16];
  sys_part_1 += 6*c3p2*c7p2*ss[20];
  sys_part_1 += 4*c3p3*c5p1*ss[14];
  sys_part_1 += 4*c3p3*c7p1*ss[16];
  sys_part_1 += 4*c5p1*c7p3*ss[26];
  sys_part_1 += 6*c5p2*c7p2*ss[24];
  sys_part_1 += 4*c5p3*c7p1*ss[22];
  sys_part_1 += c1p4*ss[4];
  sys_part_1 += c3p4*ss[12];
  sys_part_1 += c5p4*ss[20];
  sys_part_1 += c7p4*ss[28];
  sys_part_1 += -tmom[4];
  T c1p5 = c1p4*c1p1;
  T c1p6 = c1p5*c1p1;
  T c3p5 = c3p4*c3p1;
  T c3p6 = c3p5*c3p1;
  T c5p5 = c5p4*c5p1;
  T c5p6 = c5p5*c5p1;
  T c7p5 = c7p4*c7p1;
  T c7p6 = c7p5*c7p1;
  TACC sys_part_2 = convert<TACC>(120*c1p1*c3p1*c5p1*c7p3*ss[30]);
  sys_part_2 += 180*c1p1*c3p1*c5p2*c7p2*ss[28];
  sys_part_2 += 120*c1p1*c3p1*c5p3*c7p1*ss[26];
  sys_part_2 += 180*c1p1*c3p2*c5p1*c7p2*ss[26];
  sys_part_2 += 180*c1p1*c3p2*c5p2*c7p1*ss[24];
  sys_part_2 += 120*c1p1*c3p3*c5p1*c7p1*ss[22];
  sys_part_2 += 180*c1p2*c3p1*c5p1*c7p2*ss[24];
  sys_part_2 += 180*c1p2*c3p1*c5p2*c7p1*ss[22];
  sys_part_2 += 180*c1p2*c3p2*c5p1*c7p1*ss[20];
  sys_part_2 += 120*c1p3*c3p1*c5p1*c7p1*ss[18];
  sys_part_2 += 30*c1p1*c3p1*c5p4*ss[24];
  sys_part_2 += 30*c1p1*c3p1*c7p4*ss[32];
  sys_part_2 += 60*c1p1*c3p2*c5p3*ss[22];
  sys_part_2 += 60*c1p1*c3p2*c7p3*ss[28];
  sys_part_2 += 60*c1p1*c3p3*c5p2*ss[20];
  sys_part_2 += 60*c1p1*c3p3*c7p2*ss[24];
  sys_part_2 += 30*c1p1*c3p4*c5p1*ss[18];
  sys_part_2 += 30*c1p1*c3p4*c7p1*ss[20];
  sys_part_2 += 30*c1p1*c5p1*c7p4*ss[34];
  sys_part_2 += 60*c1p1*c5p2*c7p3*ss[32];
  sys_part_2 += 60*c1p1*c5p3*c7p2*ss[30];
  sys_part_2 += 30*c1p1*c5p4*c7p1*ss[28];
  sys_part_2 += 60*c1p2*c3p1*c5p3*ss[20];
  sys_part_2 += 60*c1p2*c3p1*c7p3*ss[26];
  sys_part_2 += 90*c1p2*c3p2*c5p2*ss[18];
  sys_part_2 += 90*c1p2*c3p2*c7p2*ss[22];
  sys_part_2 += 60*c1p2*c3p3*c5p1*ss[16];
  sys_part_2 += 60*c1p2*c3p3*c7p1*ss[18];
  sys_part_2 += 60*c1p2*c5p1*c7p3*ss[28];
  sys_part_2 += 90*c1p2*c5p2*c7p2*ss[26];
  sys_part_2 += 60*c1p2*c5p3*c7p1*ss[24];
  sys_part_2 += 60*c1p3*c3p1*c5p2*ss[16];
  sys_part_2 += 60*c1p3*c3p1*c7p2*ss[20];
  sys_part_2 += 60*c1p3*c3p2*c5p1*ss[14];
  sys_part_2 += 60*c1p3*c3p2*c7p1*ss[16];
  sys_part_2 += 60*c1p3*c5p1*c7p2*ss[22];
  sys_part_2 += 60*c1p3*c5p2*c7p1*ss[20];
  sys_part_2 += 30*c1p4*c3p1*c5p1*ss[12];
  sys_part_2 += 30*c1p4*c3p1*c7p1*ss[14];
  sys_part_2 += 30*c1p4*c5p1*c7p1*ss[16];
  sys_part_2 += 30*c3p1*c5p1*c7p4*ss[36];
  sys_part_2 += 60*c3p1*c5p2*c7p3*ss[34];
  sys_part_2 += 60*c3p1*c5p3*c7p2*ss[32];
  sys_part_2 += 30*c3p1*c5p4*c7p1*ss[30];
  sys_part_2 += 60*c3p2*c5p1*c7p3*ss[32];
  sys_part_2 += 90*c3p2*c5p2*c7p2*ss[30];
  sys_part_2 += 60*c3p2*c5p3*c7p1*ss[28];
  sys_part_2 += 60*c3p3*c5p1*c7p2*ss[28];
  sys_part_2 += 60*c3p3*c5p2*c7p1*ss[26];
  sys_part_2 += 30*c3p4*c5p1*c7p1*ss[24];
  sys_part_2 += 6*c1p1*c3p5*ss[16];
  sys_part_2 += 6*c1p1*c5p5*ss[26];
  sys_part_2 += 6*c1p1*c7p5*ss[36];
  sys_part_2 += 15*c1p2*c3p4*ss[14];
  sys_part_2 += 15*c1p2*c5p4*ss[22];
  sys_part_2 += 15*c1p2*c7p4*ss[30];
  sys_part_2 += 20*c1p3*c3p3*ss[12];
  sys_part_2 += 20*c1p3*c5p3*ss[18];
  sys_part_2 += 20*c1p3*c7p3*ss[24];
  sys_part_2 += 15*c1p4*c3p2*ss[10];
  sys_part_2 += 15*c1p4*c5p2*ss[14];
  sys_part_2 += 15*c1p4*c7p2*ss[18];
  sys_part_2 += 6*c1p5*c3p1*ss[8];
  sys_part_2 += 6*c1p5*c5p1*ss[10];
  sys_part_2 += 6*c1p5*c7p1*ss[12];
  sys_part_2 += 6*c3p1*c5p5*ss[28];
  sys_part_2 += 6*c3p1*c7p5*ss[38];
  sys_part_2 += 15*c3p2*c5p4*ss[26];
  sys_part_2 += 15*c3p2*c7p4*ss[34];
  sys_part_2 += 20*c3p3*c5p3*ss[24];
  sys_part_2 += 20*c3p3*c7p3*ss[30];
  sys_part_2 += 15*c3p4*c5p2*ss[22];
  sys_part_2 += 15*c3p4*c7p2*ss[26];
  sys_part_2 += 6*c3p5*c5p1*ss[20];
  sys_part_2 += 6*c3p5*c7p1*ss[22];
  sys_part_2 += 6*c5p1*c7p5*ss[40];
  sys_part_2 += 15*c5p2*c7p4*ss[38];
  sys_part_2 += 20*c5p3*c7p3*ss[36];
  sys_part_2 += 15*c5p4*c7p2*ss[34];
  sys_part_2 += 6*c5p5*c7p1*ss[32];
  sys_part_2 += c1p6*ss[6];
  sys_part_2 += c3p6*ss[18];
  sys_part_2 += c5p6*ss[30];
  sys_part_2 += c7p6*ss[42];
  sys_part_2 += -tmom[6];
  T c1p7 = c1p6*c1p1;
  T c1p8 = c1p7*c1p1;
  T c3p7 = c3p6*c3p1;
  T c3p8 = c3p7*c3p1;
  T c5p7 = c5p6*c5p1;
  T c5p8 = c5p7*c5p1;
  T c7p7 = c7p6*c7p1;
  T c7p8 = c7p7*c7p1;
  TACC sys_part_3 = convert<TACC>(336*c1p1*c3p1*c5p1*c7p5*ss[44]);
  sys_part_3 += 840*c1p1*c3p1*c5p2*c7p4*ss[42];
  sys_part_3 += 1120*c1p1*c3p1*c5p3*c7p3*ss[40];
  sys_part_3 += 840*c1p1*c3p1*c5p4*c7p2*ss[38];
  sys_part_3 += 336*c1p1*c3p1*c5p5*c7p1*ss[36];
  sys_part_3 += 840*c1p1*c3p2*c5p1*c7p4*ss[40];
  sys_part_3 += 1680*c1p1*c3p2*c5p2*c7p3*ss[38];
  sys_part_3 += 1680*c1p1*c3p2*c5p3*c7p2*ss[36];
  sys_part_3 += 840*c1p1*c3p2*c5p4*c7p1*ss[34];
  sys_part_3 += 1120*c1p1*c3p3*c5p1*c7p3*ss[36];
  sys_part_3 += 1680*c1p1*c3p3*c5p2*c7p2*ss[34];
  sys_part_3 += 1120*c1p1*c3p3*c5p3*c7p1*ss[32];
  sys_part_3 += 840*c1p1*c3p4*c5p1*c7p2*ss[32];
  sys_part_3 += 840*c1p1*c3p4*c5p2*c7p1*ss[30];
  sys_part_3 += 336*c1p1*c3p5*c5p1*c7p1*ss[28];
  sys_part_3 += 840*c1p2*c3p1*c5p1*c7p4*ss[38];
  sys_part_3 += 1680*c1p2*c3p1*c5p2*c7p3*ss[36];
  sys_part_3 += 1680*c1p2*c3p1*c5p3*c7p2*ss[34];
  sys_part_3 += 840*c1p2*c3p1*c5p4*c7p1*ss[32];
  sys_part_3 += 1680*c1p2*c3p2*c5p1*c7p3*ss[34];
  sys_part_3 += 2520*c1p2*c3p2*c5p2*c7p2*ss[32];
  sys_part_3 += 1680*c1p2*c3p2*c5p3*c7p1*ss[30];
  sys_part_3 += 1680*c1p2*c3p3*c5p1*c7p2*ss[30];
  sys_part_3 += 1680*c1p2*c3p3*c5p2*c7p1*ss[28];
  sys_part_3 += 840*c1p2*c3p4*c5p1*c7p1*ss[26];
  sys_part_3 += 1120*c1p3*c3p1*c5p1*c7p3*ss[32];
  sys_part_3 += 1680*c1p3*c3p1*c5p2*c7p2*ss[30];
  sys_part_3 += 1120*c1p3*c3p1*c5p3*c7p1*ss[28];
  sys_part_3 += 1680*c1p3*c3p2*c5p1*c7p2*ss[28];
  sys_part_3 += 1680*c1p3*c3p2*c5p2*c7p1*ss[26];
  sys_part_3 += 1120*c1p3*c3p3*c5p1*c7p1*ss[24];
  sys_part_3 += 840*c1p4*c3p1*c5p1*c7p2*ss[26];
  sys_part_3 += 840*c1p4*c3p1*c5p2*c7p1*ss[24];
  sys_part_3 += 840*c1p4*c3p2*c5p1*c7p1*ss[22];
  sys_part_3 += 336*c1p5*c3p1*c5p1*c7p1*ss[20];
  sys_part_3 += 56*c1p1*c3p1*c5p6*ss[34];
  sys_part_3 += 56*c1p1*c3p1*c7p6*ss[46];
  sys_part_3 += 168*c1p1*c3p2*c5p5*ss[32];
  sys_part_3 += 168*c1p1*c3p2*c7p5*ss[42];
  sys_part_3 += 280*c1p1*c3p3*c5p4*ss[30];
  sys_part_3 += 280*c1p1*c3p3*c7p4*ss[38];
  sys_part_3 += 280*c1p1*c3p4*c5p3*ss[28];
  sys_part_3 += 280*c1p1*c3p4*c7p3*ss[34];
  sys_part_3 += 168*c1p1*c3p5*c5p2*ss[26];
  sys_part_3 += 168*c1p1*c3p5*c7p2*ss[30];
  sys_part_3 += 56*c1p1*c3p6*c5p1*ss[24];
  sys_part_3 += 56*c1p1*c3p6*c7p1*ss[26];
  sys_part_3 += 56*c1p1*c5p1*c7p6*ss[48];
  sys_part_3 += 168*c1p1*c5p2*c7p5*ss[46];
  sys_part_3 += 280*c1p1*c5p3*c7p4*ss[44];
  sys_part_3 += 280*c1p1*c5p4*c7p3*ss[42];
  sys_part_3 += 168*c1p1*c5p5*c7p2*ss[40];
  sys_part_3 += 56*c1p1*c5p6*c7p1*ss[38];
  sys_part_3 += 168*c1p2*c3p1*c5p5*ss[30];
  sys_part_3 += 168*c1p2*c3p1*c7p5*ss[40];
  sys_part_3 += 420*c1p2*c3p2*c5p4*ss[28];
  sys_part_3 += 420*c1p2*c3p2*c7p4*ss[36];
  sys_part_3 += 560*c1p2*c3p3*c5p3*ss[26];
  sys_part_3 += 560*c1p2*c3p3*c7p3*ss[32];
  sys_part_3 += 420*c1p2*c3p4*c5p2*ss[24];
  sys_part_3 += 420*c1p2*c3p4*c7p2*ss[28];
  sys_part_3 += 168*c1p2*c3p5*c5p1*ss[22];
  sys_part_3 += 168*c1p2*c3p5*c7p1*ss[24];
  sys_part_3 += 168*c1p2*c5p1*c7p5*ss[42];
  sys_part_3 += 420*c1p2*c5p2*c7p4*ss[40];
  sys_part_3 += 560*c1p2*c5p3*c7p3*ss[38];
  sys_part_3 += 420*c1p2*c5p4*c7p2*ss[36];
  sys_part_3 += 168*c1p2*c5p5*c7p1*ss[34];
  sys_part_3 += 280*c1p3*c3p1*c5p4*ss[26];
  sys_part_3 += 280*c1p3*c3p1*c7p4*ss[34];
  sys_part_3 += 560*c1p3*c3p2*c5p3*ss[24];
  sys_part_3 += 560*c1p3*c3p2*c7p3*ss[30];
  sys_part_3 += 560*c1p3*c3p3*c5p2*ss[22];
  sys_part_3 += 560*c1p3*c3p3*c7p2*ss[26];
  sys_part_3 += 280*c1p3*c3p4*c5p1*ss[20];
  sys_part_3 += 280*c1p3*c3p4*c7p1*ss[22];
  sys_part_3 += 280*c1p3*c5p1*c7p4*ss[36];
  sys_part_3 += 560*c1p3*c5p2*c7p3*ss[34];
  sys_part_3 += 560*c1p3*c5p3*c7p2*ss[32];
  sys_part_3 += 280*c1p3*c5p4*c7p1*ss[30];
  sys_part_3 += 280*c1p4*c3p1*c5p3*ss[22];
  sys_part_3 += 280*c1p4*c3p1*c7p3*ss[28];
  sys_part_3 += 420*c1p4*c3p2*c5p2*ss[20];
  sys_part_3 += 420*c1p4*c3p2*c7p2*ss[24];
  sys_part_3 += 280*c1p4*c3p3*c5p1*ss[18];
  sys_part_3 += 280*c1p4*c3p3*c7p1*ss[20];
  sys_part_3 += 280*c1p4*c5p1*c7p3*ss[30];
  sys_part_3 += 420*c1p4*c5p2*c7p2*ss[28];
  sys_part_3 += 280*c1p4*c5p3*c7p1*ss[26];
  sys_part_3 += 168*c1p5*c3p1*c5p2*ss[18];
  sys_part_3 += 168*c1p5*c3p1*c7p2*ss[22];
  sys_part_3 += 168*c1p5*c3p2*c5p1*ss[16];
  sys_part_3 += 168*c1p5*c3p2*c7p1*ss[18];
  sys_part_3 += 168*c1p5*c5p1*c7p2*ss[24];
  sys_part_3 += 168*c1p5*c5p2*c7p1*ss[22];
  sys_part_3 += 56*c1p6*c3p1*c5p1*ss[14];
  sys_part_3 += 56*c1p6*c3p1*c7p1*ss[16];
  sys_part_3 += 56*c1p6*c5p1*c7p1*ss[18];
  sys_part_3 += 56*c3p1*c5p1*c7p6*ss[50];
  sys_part_3 += 168*c3p1*c5p2*c7p5*ss[48];
  sys_part_3 += 280*c3p1*c5p3*c7p4*ss[46];
  sys_part_3 += 280*c3p1*c5p4*c7p3*ss[44];
  sys_part_3 += 168*c3p1*c5p5*c7p2*ss[42];
  sys_part_3 += 56*c3p1*c5p6*c7p1*ss[40];
  sys_part_3 += 168*c3p2*c5p1*c7p5*ss[46];
  sys_part_3 += 420*c3p2*c5p2*c7p4*ss[44];
  sys_part_3 += 560*c3p2*c5p3*c7p3*ss[42];
  sys_part_3 += 420*c3p2*c5p4*c7p2*ss[40];
  sys_part_3 += 168*c3p2*c5p5*c7p1*ss[38];
  sys_part_3 += 280*c3p3*c5p1*c7p4*ss[42];
  sys_part_3 += 560*c3p3*c5p2*c7p3*ss[40];
  sys_part_3 += 560*c3p3*c5p3*c7p2*ss[38];
  sys_part_3 += 280*c3p3*c5p4*c7p1*ss[36];
  sys_part_3 += 280*c3p4*c5p1*c7p3*ss[38];
  sys_part_3 += 420*c3p4*c5p2*c7p2*ss[36];
  sys_part_3 += 280*c3p4*c5p3*c7p1*ss[34];
  sys_part_3 += 168*c3p5*c5p1*c7p2*ss[34];
  sys_part_3 += 168*c3p5*c5p2*c7p1*ss[32];
  sys_part_3 += 56*c3p6*c5p1*c7p1*ss[30];
  sys_part_3 += 8*c1p1*c3p7*ss[22];
  sys_part_3 += 8*c1p1*c5p7*ss[36];
  sys_part_3 += 8*c1p1*c7p7*ss[50];
  sys_part_3 += 28*c1p2*c3p6*ss[20];
  sys_part_3 += 28*c1p2*c5p6*ss[32];
  sys_part_3 += 28*c1p2*c7p6*ss[44];
  sys_part_3 += 56*c1p3*c3p5*ss[18];
  sys_part_3 += 56*c1p3*c5p5*ss[28];
  sys_part_3 += 56*c1p3*c7p5*ss[38];
  sys_part_3 += 70*c1p4*c3p4*ss[16];
  sys_part_3 += 70*c1p4*c5p4*ss[24];
  sys_part_3 += 70*c1p4*c7p4*ss[32];
  sys_part_3 += 56*c1p5*c3p3*ss[14];
  sys_part_3 += 56*c1p5*c5p3*ss[20];
  sys_part_3 += 56*c1p5*c7p3*ss[26];
  sys_part_3 += 28*c1p6*c3p2*ss[12];
  sys_part_3 += 28*c1p6*c5p2*ss[16];
  sys_part_3 += 28*c1p6*c7p2*ss[20];
  sys_part_3 += 8*c1p7*c3p1*ss[10];
  sys_part_3 += 8*c1p7*c5p1*ss[12];
  sys_part_3 += 8*c1p7*c7p1*ss[14];
  sys_part_3 += 8*c3p1*c5p7*ss[38];
  sys_part_3 += 8*c3p1*c7p7*ss[52];
  sys_part_3 += 28*c3p2*c5p6*ss[36];
  sys_part_3 += 28*c3p2*c7p6*ss[48];
  sys_part_3 += 56*c3p3*c5p5*ss[34];
  sys_part_3 += 56*c3p3*c7p5*ss[44];
  sys_part_3 += 70*c3p4*c5p4*ss[32];
  sys_part_3 += 70*c3p4*c7p4*ss[40];
  sys_part_3 += 56*c3p5*c5p3*ss[30];
  sys_part_3 += 56*c3p5*c7p3*ss[36];
  sys_part_3 += 28*c3p6*c5p2*ss[28];
  sys_part_3 += 28*c3p6*c7p2*ss[32];
  sys_part_3 += 8*c3p7*c5p1*ss[26];
  sys_part_3 += 8*c3p7*c7p1*ss[28];
  sys_part_3 += 8*c5p1*c7p7*ss[54];
  sys_part_3 += 28*c5p2*c7p6*ss[52];
  sys_part_3 += 56*c5p3*c7p5*ss[50];
  sys_part_3 += 70*c5p4*c7p4*ss[48];
  sys_part_3 += 56*c5p5*c7p3*ss[46];
  sys_part_3 += 28*c5p6*c7p2*ss[44];
  sys_part_3 += 8*c5p7*c7p1*ss[42];
  sys_part_3 += c1p8*ss[8];
  sys_part_3 += c3p8*ss[24];
  sys_part_3 += c5p8*ss[40];
  sys_part_3 += c7p8*ss[56];
  sys_part_3 += -tmom[8];
  TACC grad_0_num = convert<TACC>(2*(c1p1*ss[2] + c3p1*ss[4] + c5p1*ss[6] + c7p1*ss[8])*sys_part_0);
  grad_0_num += 4*(6*c1p1*c3p1*c5p1*ss[10] + 6*c1p1*c3p1*c7p1*ss[12] + 6*c1p1*c5p1*c7p1*ss[14] + 6*c3p1*c5p1*c7p1*ss[16] + 3*c1p1*c3p2*ss[8] + 3*c1p1*c5p2*ss[12] + 3*c1p1*c7p2*ss[16] + 3*c1p2*c3p1*ss[6] + 3*c1p2*c5p1*ss[8] + 3*c1p2*c7p1*ss[10] + 3*c3p1*c5p2*ss[14] + 3*c3p1*c7p2*ss[18] + 3*c3p2*c5p1*ss[12] + 3*c3p2*c7p1*ss[14] + 3*c5p1*c7p2*ss[20] + 3*c5p2*c7p1*ss[18] + c1p3*ss[4] + c3p3*ss[10] + c5p3*ss[16] + c7p3*ss[22])*sys_part_1;
  grad_0_num += 6*(60*c1p1*c3p1*c5p1*c7p2*ss[24] + 60*c1p1*c3p1*c5p2*c7p1*ss[22] + 60*c1p1*c3p2*c5p1*c7p1*ss[20] + 60*c1p2*c3p1*c5p1*c7p1*ss[18] + 20*c1p1*c3p1*c5p3*ss[20] + 20*c1p1*c3p1*c7p3*ss[26] + 30*c1p1*c3p2*c5p2*ss[18] + 30*c1p1*c3p2*c7p2*ss[22] + 20*c1p1*c3p3*c5p1*ss[16] + 20*c1p1*c3p3*c7p1*ss[18] + 20*c1p1*c5p1*c7p3*ss[28] + 30*c1p1*c5p2*c7p2*ss[26] + 20*c1p1*c5p3*c7p1*ss[24] + 30*c1p2*c3p1*c5p2*ss[16] + 30*c1p2*c3p1*c7p2*ss[20] + 30*c1p2*c3p2*c5p1*ss[14] + 30*c1p2*c3p2*c7p1*ss[16] + 30*c1p2*c5p1*c7p2*ss[22] + 30*c1p2*c5p2*c7p1*ss[20] + 20*c1p3*c3p1*c5p1*ss[12] + 20*c1p3*c3p1*c7p1*ss[14] + 20*c1p3*c5p1*c7p1*ss[16] + 20*c3p1*c5p1*c7p3*ss[30] + 30*c3p1*c5p2*c7p2*ss[28] + 20*c3p1*c5p3*c7p1*ss[26] + 30*c3p2*c5p1*c7p2*ss[26] + 30*c3p2*c5p2*c7p1*ss[24] + 20*c3p3*c5p1*c7p1*ss[22] + 5*c1p1*c3p4*ss[14] + 5*c1p1*c5p4*ss[22] + 5*c1p1*c7p4*ss[30] + 10*c1p2*c3p3*ss[12] + 10*c1p2*c5p3*ss[18] + 10*c1p2*c7p3*ss[24] + 10*c1p3*c3p2*ss[10] + 10*c1p3*c5p2*ss[14] + 10*c1p3*c7p2*ss[18] + 5*c1p4*c3p1*ss[8] + 5*c1p4*c5p1*ss[10] + 5*c1p4*c7p1*ss[12] + 5*c3p1*c5p4*ss[24] + 5*c3p1*c7p4*ss[32] + 10*c3p2*c5p3*ss[22] + 10*c3p2*c7p3*ss[28] + 10*c3p3*c5p2*ss[20] + 10*c3p3*c7p2*ss[24] + 5*c3p4*c5p1*ss[18] + 5*c3p4*c7p1*ss[20] + 5*c5p1*c7p4*ss[34] + 10*c5p2*c7p3*ss[32] + 10*c5p3*c7p2*ss[30] + 5*c5p4*c7p1*ss[28] + c1p5*ss[6] + c3p5*ss[16] + c5p5*ss[26] + c7p5*ss[36])*sys_part_2;
  grad_0_num += 8*(210*c1p1*c3p1*c5p1*c7p4*ss[38] + 420*c1p1*c3p1*c5p2*c7p3*ss[36] + 420*c1p1*c3p1*c5p3*c7p2*ss[34] + 210*c1p1*c3p1*c5p4*c7p1*ss[32] + 420*c1p1*c3p2*c5p1*c7p3*ss[34] + 630*c1p1*c3p2*c5p2*c7p2*ss[32] + 420*c1p1*c3p2*c5p3*c7p1*ss[30] + 420*c1p1*c3p3*c5p1*c7p2*ss[30] + 420*c1p1*c3p3*c5p2*c7p1*ss[28] + 210*c1p1*c3p4*c5p1*c7p1*ss[26] + 420*c1p2*c3p1*c5p1*c7p3*ss[32] + 630*c1p2*c3p1*c5p2*c7p2*ss[30] + 420*c1p2*c3p1*c5p3*c7p1*ss[28] + 630*c1p2*c3p2*c5p1*c7p2*ss[28] + 630*c1p2*c3p2*c5p2*c7p1*ss[26] + 420*c1p2*c3p3*c5p1*c7p1*ss[24] + 420*c1p3*c3p1*c5p1*c7p2*ss[26] + 420*c1p3*c3p1*c5p2*c7p1*ss[24] + 420*c1p3*c3p2*c5p1*c7p1*ss[22] + 210*c1p4*c3p1*c5p1*c7p1*ss[20] + 42*c1p1*c3p1*c5p5*ss[30] + 42*c1p1*c3p1*c7p5*ss[40] + 105*c1p1*c3p2*c5p4*ss[28] + 105*c1p1*c3p2*c7p4*ss[36] + 140*c1p1*c3p3*c5p3*ss[26] + 140*c1p1*c3p3*c7p3*ss[32] + 105*c1p1*c3p4*c5p2*ss[24] + 105*c1p1*c3p4*c7p2*ss[28] + 42*c1p1*c3p5*c5p1*ss[22] + 42*c1p1*c3p5*c7p1*ss[24] + 42*c1p1*c5p1*c7p5*ss[42] + 105*c1p1*c5p2*c7p4*ss[40] + 140*c1p1*c5p3*c7p3*ss[38] + 105*c1p1*c5p4*c7p2*ss[36] + 42*c1p1*c5p5*c7p1*ss[34] + 105*c1p2*c3p1*c5p4*ss[26] + 105*c1p2*c3p1*c7p4*ss[34] + 210*c1p2*c3p2*c5p3*ss[24] + 210*c1p2*c3p2*c7p3*ss[30] + 210*c1p2*c3p3*c5p2*ss[22] + 210*c1p2*c3p3*c7p2*ss[26] + 105*c1p2*c3p4*c5p1*ss[20] + 105*c1p2*c3p4*c7p1*ss[22] + 105*c1p2*c5p1*c7p4*ss[36] + 210*c1p2*c5p2*c7p3*ss[34] + 210*c1p2*c5p3*c7p2*ss[32] + 105*c1p2*c5p4*c7p1*ss[30] + 140*c1p3*c3p1*c5p3*ss[22] + 140*c1p3*c3p1*c7p3*ss[28] + 210*c1p3*c3p2*c5p2*ss[20] + 210*c1p3*c3p2*c7p2*ss[24] + 140*c1p3*c3p3*c5p1*ss[18] + 140*c1p3*c3p3*c7p1*ss[20] + 140*c1p3*c5p1*c7p3*ss[30] + 210*c1p3*c5p2*c7p2*ss[28] + 140*c1p3*c5p3*c7p1*ss[26] + 105*c1p4*c3p1*c5p2*ss[18] + 105*c1p4*c3p1*c7p2*ss[22] + 105*c1p4*c3p2*c5p1*ss[16] + 105*c1p4*c3p2*c7p1*ss[18] + 105*c1p4*c5p1*c7p2*ss[24] + 105*c1p4*c5p2*c7p1*ss[22] + 42*c1p5*c3p1*c5p1*ss[14] + 42*c1p5*c3p1*c7p1*ss[16] + 42*c1p5*c5p1*c7p1*ss[18] + 42*c3p1*c5p1*c7p5*ss[44] + 105*c3p1*c5p2*c7p4*ss[42] + 140*c3p1*c5p3*c7p3*ss[40] + 105*c3p1*c5p4*c7p2*ss[38] + 42*c3p1*c5p5*c7p1*ss[36] + 105*c3p2*c5p1*c7p4*ss[40] + 210*c3p2*c5p2*c7p3*ss[38] + 210*c3p2*c5p3*c7p2*ss[36] + 105*c3p2*c5p4*c7p1*ss[34] + 140*c3p3*c5p1*c7p3*ss[36] + 210*c3p3*c5p2*c7p2*ss[34] + 140*c3p3*c5p3*c7p1*ss[32] + 105*c3p4*c5p1*c7p2*ss[32] + 105*c3p4*c5p2*c7p1*ss[30] + 42*c3p5*c5p1*c7p1*ss[28] + 7*c1p1*c3p6*ss[20] + 7*c1p1*c5p6*ss[32] + 7*c1p1*c7p6*ss[44] + 21*c1p2*c3p5*ss[18] + 21*c1p2*c5p5*ss[28] + 21*c1p2*c7p5*ss[38] + 35*c1p3*c3p4*ss[16] + 35*c1p3*c5p4*ss[24] + 35*c1p3*c7p4*ss[32] + 35*c1p4*c3p3*ss[14] + 35*c1p4*c5p3*ss[20] + 35*c1p4*c7p3*ss[26] + 21*c1p5*c3p2*ss[12] + 21*c1p5*c5p2*ss[16] + 21*c1p5*c7p2*ss[20] + 7*c1p6*c3p1*ss[10] + 7*c1p6*c5p1*ss[12] + 7*c1p6*c7p1*ss[14] + 7*c3p1*c5p6*ss[34] + 7*c3p1*c7p6*ss[46] + 21*c3p2*c5p5*ss[32] + 21*c3p2*c7p5*ss[42] + 35*c3p3*c5p4*ss[30] + 35*c3p3*c7p4*ss[38] + 35*c3p4*c5p3*ss[28] + 35*c3p4*c7p3*ss[34] + 21*c3p5*c5p2*ss[26] + 21*c3p5*c7p2*ss[30] + 7*c3p6*c5p1*ss[24] + 7*c3p6*c7p1*ss[26] + 7*c5p1*c7p6*ss[48] + 21*c5p2*c7p5*ss[46] + 35*c5p3*c7p4*ss[44] + 35*c5p4*c7p3*ss[42] + 21*c5p5*c7p2*ss[40] + 7*c5p6*c7p1*ss[38] + c1p7*ss[8] + c3p7*ss[22] + c5p7*ss[36] + c7p7*ss[50])*sys_part_3;
  TACC grad_0_den = sqrt(square(sys_part_0) + square(sys_part_1) + square(sys_part_2) + square(sys_part_3));
  TACC grad_1_num = convert<TACC>(2*(c1p1*ss[4] + c3p1*ss[6] + c5p1*ss[8] + c7p1*ss[10])*sys_part_0);
  grad_1_num += 4*(6*c1p1*c3p1*c5p1*ss[12] + 6*c1p1*c3p1*c7p1*ss[14] + 6*c1p1*c5p1*c7p1*ss[16] + 6*c3p1*c5p1*c7p1*ss[18] + 3*c1p1*c3p2*ss[10] + 3*c1p1*c5p2*ss[14] + 3*c1p1*c7p2*ss[18] + 3*c1p2*c3p1*ss[8] + 3*c1p2*c5p1*ss[10] + 3*c1p2*c7p1*ss[12] + 3*c3p1*c5p2*ss[16] + 3*c3p1*c7p2*ss[20] + 3*c3p2*c5p1*ss[14] + 3*c3p2*c7p1*ss[16] + 3*c5p1*c7p2*ss[22] + 3*c5p2*c7p1*ss[20] + c1p3*ss[6] + c3p3*ss[12] + c5p3*ss[18] + c7p3*ss[24])*sys_part_1;
  grad_1_num += 6*(60*c1p1*c3p1*c5p1*c7p2*ss[26] + 60*c1p1*c3p1*c5p2*c7p1*ss[24] + 60*c1p1*c3p2*c5p1*c7p1*ss[22] + 60*c1p2*c3p1*c5p1*c7p1*ss[20] + 20*c1p1*c3p1*c5p3*ss[22] + 20*c1p1*c3p1*c7p3*ss[28] + 30*c1p1*c3p2*c5p2*ss[20] + 30*c1p1*c3p2*c7p2*ss[24] + 20*c1p1*c3p3*c5p1*ss[18] + 20*c1p1*c3p3*c7p1*ss[20] + 20*c1p1*c5p1*c7p3*ss[30] + 30*c1p1*c5p2*c7p2*ss[28] + 20*c1p1*c5p3*c7p1*ss[26] + 30*c1p2*c3p1*c5p2*ss[18] + 30*c1p2*c3p1*c7p2*ss[22] + 30*c1p2*c3p2*c5p1*ss[16] + 30*c1p2*c3p2*c7p1*ss[18] + 30*c1p2*c5p1*c7p2*ss[24] + 30*c1p2*c5p2*c7p1*ss[22] + 20*c1p3*c3p1*c5p1*ss[14] + 20*c1p3*c3p1*c7p1*ss[16] + 20*c1p3*c5p1*c7p1*ss[18] + 20*c3p1*c5p1*c7p3*ss[32] + 30*c3p1*c5p2*c7p2*ss[30] + 20*c3p1*c5p3*c7p1*ss[28] + 30*c3p2*c5p1*c7p2*ss[28] + 30*c3p2*c5p2*c7p1*ss[26] + 20*c3p3*c5p1*c7p1*ss[24] + 5*c1p1*c3p4*ss[16] + 5*c1p1*c5p4*ss[24] + 5*c1p1*c7p4*ss[32] + 10*c1p2*c3p3*ss[14] + 10*c1p2*c5p3*ss[20] + 10*c1p2*c7p3*ss[26] + 10*c1p3*c3p2*ss[12] + 10*c1p3*c5p2*ss[16] + 10*c1p3*c7p2*ss[20] + 5*c1p4*c3p1*ss[10] + 5*c1p4*c5p1*ss[12] + 5*c1p4*c7p1*ss[14] + 5*c3p1*c5p4*ss[26] + 5*c3p1*c7p4*ss[34] + 10*c3p2*c5p3*ss[24] + 10*c3p2*c7p3*ss[30] + 10*c3p3*c5p2*ss[22] + 10*c3p3*c7p2*ss[26] + 5*c3p4*c5p1*ss[20] + 5*c3p4*c7p1*ss[22] + 5*c5p1*c7p4*ss[36] + 10*c5p2*c7p3*ss[34] + 10*c5p3*c7p2*ss[32] + 5*c5p4*c7p1*ss[30] + c1p5*ss[8] + c3p5*ss[18] + c5p5*ss[28] + c7p5*ss[38])*sys_part_2;
  grad_1_num += 8*(210*c1p1*c3p1*c5p1*c7p4*ss[40] + 420*c1p1*c3p1*c5p2*c7p3*ss[38] + 420*c1p1*c3p1*c5p3*c7p2*ss[36] + 210*c1p1*c3p1*c5p4*c7p1*ss[34] + 420*c1p1*c3p2*c5p1*c7p3*ss[36] + 630*c1p1*c3p2*c5p2*c7p2*ss[34] + 420*c1p1*c3p2*c5p3*c7p1*ss[32] + 420*c1p1*c3p3*c5p1*c7p2*ss[32] + 420*c1p1*c3p3*c5p2*c7p1*ss[30] + 210*c1p1*c3p4*c5p1*c7p1*ss[28] + 420*c1p2*c3p1*c5p1*c7p3*ss[34] + 630*c1p2*c3p1*c5p2*c7p2*ss[32] + 420*c1p2*c3p1*c5p3*c7p1*ss[30] + 630*c1p2*c3p2*c5p1*c7p2*ss[30] + 630*c1p2*c3p2*c5p2*c7p1*ss[28] + 420*c1p2*c3p3*c5p1*c7p1*ss[26] + 420*c1p3*c3p1*c5p1*c7p2*ss[28] + 420*c1p3*c3p1*c5p2*c7p1*ss[26] + 420*c1p3*c3p2*c5p1*c7p1*ss[24] + 210*c1p4*c3p1*c5p1*c7p1*ss[22] + 42*c1p1*c3p1*c5p5*ss[32] + 42*c1p1*c3p1*c7p5*ss[42] + 105*c1p1*c3p2*c5p4*ss[30] + 105*c1p1*c3p2*c7p4*ss[38] + 140*c1p1*c3p3*c5p3*ss[28] + 140*c1p1*c3p3*c7p3*ss[34] + 105*c1p1*c3p4*c5p2*ss[26] + 105*c1p1*c3p4*c7p2*ss[30] + 42*c1p1*c3p5*c5p1*ss[24] + 42*c1p1*c3p5*c7p1*ss[26] + 42*c1p1*c5p1*c7p5*ss[44] + 105*c1p1*c5p2*c7p4*ss[42] + 140*c1p1*c5p3*c7p3*ss[40] + 105*c1p1*c5p4*c7p2*ss[38] + 42*c1p1*c5p5*c7p1*ss[36] + 105*c1p2*c3p1*c5p4*ss[28] + 105*c1p2*c3p1*c7p4*ss[36] + 210*c1p2*c3p2*c5p3*ss[26] + 210*c1p2*c3p2*c7p3*ss[32] + 210*c1p2*c3p3*c5p2*ss[24] + 210*c1p2*c3p3*c7p2*ss[28] + 105*c1p2*c3p4*c5p1*ss[22] + 105*c1p2*c3p4*c7p1*ss[24] + 105*c1p2*c5p1*c7p4*ss[38] + 210*c1p2*c5p2*c7p3*ss[36] + 210*c1p2*c5p3*c7p2*ss[34] + 105*c1p2*c5p4*c7p1*ss[32] + 140*c1p3*c3p1*c5p3*ss[24] + 140*c1p3*c3p1*c7p3*ss[30] + 210*c1p3*c3p2*c5p2*ss[22] + 210*c1p3*c3p2*c7p2*ss[26] + 140*c1p3*c3p3*c5p1*ss[20] + 140*c1p3*c3p3*c7p1*ss[22] + 140*c1p3*c5p1*c7p3*ss[32] + 210*c1p3*c5p2*c7p2*ss[30] + 140*c1p3*c5p3*c7p1*ss[28] + 105*c1p4*c3p1*c5p2*ss[20] + 105*c1p4*c3p1*c7p2*ss[24] + 105*c1p4*c3p2*c5p1*ss[18] + 105*c1p4*c3p2*c7p1*ss[20] + 105*c1p4*c5p1*c7p2*ss[26] + 105*c1p4*c5p2*c7p1*ss[24] + 42*c1p5*c3p1*c5p1*ss[16] + 42*c1p5*c3p1*c7p1*ss[18] + 42*c1p5*c5p1*c7p1*ss[20] + 42*c3p1*c5p1*c7p5*ss[46] + 105*c3p1*c5p2*c7p4*ss[44] + 140*c3p1*c5p3*c7p3*ss[42] + 105*c3p1*c5p4*c7p2*ss[40] + 42*c3p1*c5p5*c7p1*ss[38] + 105*c3p2*c5p1*c7p4*ss[42] + 210*c3p2*c5p2*c7p3*ss[40] + 210*c3p2*c5p3*c7p2*ss[38] + 105*c3p2*c5p4*c7p1*ss[36] + 140*c3p3*c5p1*c7p3*ss[38] + 210*c3p3*c5p2*c7p2*ss[36] + 140*c3p3*c5p3*c7p1*ss[34] + 105*c3p4*c5p1*c7p2*ss[34] + 105*c3p4*c5p2*c7p1*ss[32] + 42*c3p5*c5p1*c7p1*ss[30] + 7*c1p1*c3p6*ss[22] + 7*c1p1*c5p6*ss[34] + 7*c1p1*c7p6*ss[46] + 21*c1p2*c3p5*ss[20] + 21*c1p2*c5p5*ss[30] + 21*c1p2*c7p5*ss[40] + 35*c1p3*c3p4*ss[18] + 35*c1p3*c5p4*ss[26] + 35*c1p3*c7p4*ss[34] + 35*c1p4*c3p3*ss[16] + 35*c1p4*c5p3*ss[22] + 35*c1p4*c7p3*ss[28] + 21*c1p5*c3p2*ss[14] + 21*c1p5*c5p2*ss[18] + 21*c1p5*c7p2*ss[22] + 7*c1p6*c3p1*ss[12] + 7*c1p6*c5p1*ss[14] + 7*c1p6*c7p1*ss[16] + 7*c3p1*c5p6*ss[36] + 7*c3p1*c7p6*ss[48] + 21*c3p2*c5p5*ss[34] + 21*c3p2*c7p5*ss[44] + 35*c3p3*c5p4*ss[32] + 35*c3p3*c7p4*ss[40] + 35*c3p4*c5p3*ss[30] + 35*c3p4*c7p3*ss[36] + 21*c3p5*c5p2*ss[28] + 21*c3p5*c7p2*ss[32] + 7*c3p6*c5p1*ss[26] + 7*c3p6*c7p1*ss[28] + 7*c5p1*c7p6*ss[50] + 21*c5p2*c7p5*ss[48] + 35*c5p3*c7p4*ss[46] + 35*c5p4*c7p3*ss[44] + 21*c5p5*c7p2*ss[42] + 7*c5p6*c7p1*ss[40] + c1p7*ss[10] + c3p7*ss[24] + c5p7*ss[38] + c7p7*ss[52])*sys_part_3;
  TACC grad_1_den = sqrt(square(sys_part_0) + square(sys_part_1) + square(sys_part_2) + square(sys_part_3));
  TACC grad_2_num = convert<TACC>(2*(c1p1*ss[6] + c3p1*ss[8] + c5p1*ss[10] + c7p1*ss[12])*sys_part_0);
  grad_2_num += 4*(6*c1p1*c3p1*c5p1*ss[14] + 6*c1p1*c3p1*c7p1*ss[16] + 6*c1p1*c5p1*c7p1*ss[18] + 6*c3p1*c5p1*c7p1*ss[20] + 3*c1p1*c3p2*ss[12] + 3*c1p1*c5p2*ss[16] + 3*c1p1*c7p2*ss[20] + 3*c1p2*c3p1*ss[10] + 3*c1p2*c5p1*ss[12] + 3*c1p2*c7p1*ss[14] + 3*c3p1*c5p2*ss[18] + 3*c3p1*c7p2*ss[22] + 3*c3p2*c5p1*ss[16] + 3*c3p2*c7p1*ss[18] + 3*c5p1*c7p2*ss[24] + 3*c5p2*c7p1*ss[22] + c1p3*ss[8] + c3p3*ss[14] + c5p3*ss[20] + c7p3*ss[26])*sys_part_1;
  grad_2_num += 6*(60*c1p1*c3p1*c5p1*c7p2*ss[28] + 60*c1p1*c3p1*c5p2*c7p1*ss[26] + 60*c1p1*c3p2*c5p1*c7p1*ss[24] + 60*c1p2*c3p1*c5p1*c7p1*ss[22] + 20*c1p1*c3p1*c5p3*ss[24] + 20*c1p1*c3p1*c7p3*ss[30] + 30*c1p1*c3p2*c5p2*ss[22] + 30*c1p1*c3p2*c7p2*ss[26] + 20*c1p1*c3p3*c5p1*ss[20] + 20*c1p1*c3p3*c7p1*ss[22] + 20*c1p1*c5p1*c7p3*ss[32] + 30*c1p1*c5p2*c7p2*ss[30] + 20*c1p1*c5p3*c7p1*ss[28] + 30*c1p2*c3p1*c5p2*ss[20] + 30*c1p2*c3p1*c7p2*ss[24] + 30*c1p2*c3p2*c5p1*ss[18] + 30*c1p2*c3p2*c7p1*ss[20] + 30*c1p2*c5p1*c7p2*ss[26] + 30*c1p2*c5p2*c7p1*ss[24] + 20*c1p3*c3p1*c5p1*ss[16] + 20*c1p3*c3p1*c7p1*ss[18] + 20*c1p3*c5p1*c7p1*ss[20] + 20*c3p1*c5p1*c7p3*ss[34] + 30*c3p1*c5p2*c7p2*ss[32] + 20*c3p1*c5p3*c7p1*ss[30] + 30*c3p2*c5p1*c7p2*ss[30] + 30*c3p2*c5p2*c7p1*ss[28] + 20*c3p3*c5p1*c7p1*ss[26] + 5*c1p1*c3p4*ss[18] + 5*c1p1*c5p4*ss[26] + 5*c1p1*c7p4*ss[34] + 10*c1p2*c3p3*ss[16] + 10*c1p2*c5p3*ss[22] + 10*c1p2*c7p3*ss[28] + 10*c1p3*c3p2*ss[14] + 10*c1p3*c5p2*ss[18] + 10*c1p3*c7p2*ss[22] + 5*c1p4*c3p1*ss[12] + 5*c1p4*c5p1*ss[14] + 5*c1p4*c7p1*ss[16] + 5*c3p1*c5p4*ss[28] + 5*c3p1*c7p4*ss[36] + 10*c3p2*c5p3*ss[26] + 10*c3p2*c7p3*ss[32] + 10*c3p3*c5p2*ss[24] + 10*c3p3*c7p2*ss[28] + 5*c3p4*c5p1*ss[22] + 5*c3p4*c7p1*ss[24] + 5*c5p1*c7p4*ss[38] + 10*c5p2*c7p3*ss[36] + 10*c5p3*c7p2*ss[34] + 5*c5p4*c7p1*ss[32] + c1p5*ss[10] + c3p5*ss[20] + c5p5*ss[30] + c7p5*ss[40])*sys_part_2;
  grad_2_num += 8*(210*c1p1*c3p1*c5p1*c7p4*ss[42] + 420*c1p1*c3p1*c5p2*c7p3*ss[40] + 420*c1p1*c3p1*c5p3*c7p2*ss[38] + 210*c1p1*c3p1*c5p4*c7p1*ss[36] + 420*c1p1*c3p2*c5p1*c7p3*ss[38] + 630*c1p1*c3p2*c5p2*c7p2*ss[36] + 420*c1p1*c3p2*c5p3*c7p1*ss[34] + 420*c1p1*c3p3*c5p1*c7p2*ss[34] + 420*c1p1*c3p3*c5p2*c7p1*ss[32] + 210*c1p1*c3p4*c5p1*c7p1*ss[30] + 420*c1p2*c3p1*c5p1*c7p3*ss[36] + 630*c1p2*c3p1*c5p2*c7p2*ss[34] + 420*c1p2*c3p1*c5p3*c7p1*ss[32] + 630*c1p2*c3p2*c5p1*c7p2*ss[32] + 630*c1p2*c3p2*c5p2*c7p1*ss[30] + 420*c1p2*c3p3*c5p1*c7p1*ss[28] + 420*c1p3*c3p1*c5p1*c7p2*ss[30] + 420*c1p3*c3p1*c5p2*c7p1*ss[28] + 420*c1p3*c3p2*c5p1*c7p1*ss[26] + 210*c1p4*c3p1*c5p1*c7p1*ss[24] + 42*c1p1*c3p1*c5p5*ss[34] + 42*c1p1*c3p1*c7p5*ss[44] + 105*c1p1*c3p2*c5p4*ss[32] + 105*c1p1*c3p2*c7p4*ss[40] + 140*c1p1*c3p3*c5p3*ss[30] + 140*c1p1*c3p3*c7p3*ss[36] + 105*c1p1*c3p4*c5p2*ss[28] + 105*c1p1*c3p4*c7p2*ss[32] + 42*c1p1*c3p5*c5p1*ss[26] + 42*c1p1*c3p5*c7p1*ss[28] + 42*c1p1*c5p1*c7p5*ss[46] + 105*c1p1*c5p2*c7p4*ss[44] + 140*c1p1*c5p3*c7p3*ss[42] + 105*c1p1*c5p4*c7p2*ss[40] + 42*c1p1*c5p5*c7p1*ss[38] + 105*c1p2*c3p1*c5p4*ss[30] + 105*c1p2*c3p1*c7p4*ss[38] + 210*c1p2*c3p2*c5p3*ss[28] + 210*c1p2*c3p2*c7p3*ss[34] + 210*c1p2*c3p3*c5p2*ss[26] + 210*c1p2*c3p3*c7p2*ss[30] + 105*c1p2*c3p4*c5p1*ss[24] + 105*c1p2*c3p4*c7p1*ss[26] + 105*c1p2*c5p1*c7p4*ss[40] + 210*c1p2*c5p2*c7p3*ss[38] + 210*c1p2*c5p3*c7p2*ss[36] + 105*c1p2*c5p4*c7p1*ss[34] + 140*c1p3*c3p1*c5p3*ss[26] + 140*c1p3*c3p1*c7p3*ss[32] + 210*c1p3*c3p2*c5p2*ss[24] + 210*c1p3*c3p2*c7p2*ss[28] + 140*c1p3*c3p3*c5p1*ss[22] + 140*c1p3*c3p3*c7p1*ss[24] + 140*c1p3*c5p1*c7p3*ss[34] + 210*c1p3*c5p2*c7p2*ss[32] + 140*c1p3*c5p3*c7p1*ss[30] + 105*c1p4*c3p1*c5p2*ss[22] + 105*c1p4*c3p1*c7p2*ss[26] + 105*c1p4*c3p2*c5p1*ss[20] + 105*c1p4*c3p2*c7p1*ss[22] + 105*c1p4*c5p1*c7p2*ss[28] + 105*c1p4*c5p2*c7p1*ss[26] + 42*c1p5*c3p1*c5p1*ss[18] + 42*c1p5*c3p1*c7p1*ss[20] + 42*c1p5*c5p1*c7p1*ss[22] + 42*c3p1*c5p1*c7p5*ss[48] + 105*c3p1*c5p2*c7p4*ss[46] + 140*c3p1*c5p3*c7p3*ss[44] + 105*c3p1*c5p4*c7p2*ss[42] + 42*c3p1*c5p5*c7p1*ss[40] + 105*c3p2*c5p1*c7p4*ss[44] + 210*c3p2*c5p2*c7p3*ss[42] + 210*c3p2*c5p3*c7p2*ss[40] + 105*c3p2*c5p4*c7p1*ss[38] + 140*c3p3*c5p1*c7p3*ss[40] + 210*c3p3*c5p2*c7p2*ss[38] + 140*c3p3*c5p3*c7p1*ss[36] + 105*c3p4*c5p1*c7p2*ss[36] + 105*c3p4*c5p2*c7p1*ss[34] + 42*c3p5*c5p1*c7p1*ss[32] + 7*c1p1*c3p6*ss[24] + 7*c1p1*c5p6*ss[36] + 7*c1p1*c7p6*ss[48] + 21*c1p2*c3p5*ss[22] + 21*c1p2*c5p5*ss[32] + 21*c1p2*c7p5*ss[42] + 35*c1p3*c3p4*ss[20] + 35*c1p3*c5p4*ss[28] + 35*c1p3*c7p4*ss[36] + 35*c1p4*c3p3*ss[18] + 35*c1p4*c5p3*ss[24] + 35*c1p4*c7p3*ss[30] + 21*c1p5*c3p2*ss[16] + 21*c1p5*c5p2*ss[20] + 21*c1p5*c7p2*ss[24] + 7*c1p6*c3p1*ss[14] + 7*c1p6*c5p1*ss[16] + 7*c1p6*c7p1*ss[18] + 7*c3p1*c5p6*ss[38] + 7*c3p1*c7p6*ss[50] + 21*c3p2*c5p5*ss[36] + 21*c3p2*c7p5*ss[46] + 35*c3p3*c5p4*ss[34] + 35*c3p3*c7p4*ss[42] + 35*c3p4*c5p3*ss[32] + 35*c3p4*c7p3*ss[38] + 21*c3p5*c5p2*ss[30] + 21*c3p5*c7p2*ss[34] + 7*c3p6*c5p1*ss[28] + 7*c3p6*c7p1*ss[30] + 7*c5p1*c7p6*ss[52] + 21*c5p2*c7p5*ss[50] + 35*c5p3*c7p4*ss[48] + 35*c5p4*c7p3*ss[46] + 21*c5p5*c7p2*ss[44] + 7*c5p6*c7p1*ss[42] + c1p7*ss[12] + c3p7*ss[26] + c5p7*ss[40] + c7p7*ss[54])*sys_part_3;
  TACC grad_2_den = sqrt(square(sys_part_0) + square(sys_part_1) + square(sys_part_2) + square(sys_part_3));
  TACC grad_3_num = convert<TACC>(2*(c1p1*ss[8] + c3p1*ss[10] + c5p1*ss[12] + c7p1*ss[14])*sys_part_0);
  grad_3_num += 4*(6*c1p1*c3p1*c5p1*ss[16] + 6*c1p1*c3p1*c7p1*ss[18] + 6*c1p1*c5p1*c7p1*ss[20] + 6*c3p1*c5p1*c7p1*ss[22] + 3*c1p1*c3p2*ss[14] + 3*c1p1*c5p2*ss[18] + 3*c1p1*c7p2*ss[22] + 3*c1p2*c3p1*ss[12] + 3*c1p2*c5p1*ss[14] + 3*c1p2*c7p1*ss[16] + 3*c3p1*c5p2*ss[20] + 3*c3p1*c7p2*ss[24] + 3*c3p2*c5p1*ss[18] + 3*c3p2*c7p1*ss[20] + 3*c5p1*c7p2*ss[26] + 3*c5p2*c7p1*ss[24] + c1p3*ss[10] + c3p3*ss[16] + c5p3*ss[22] + c7p3*ss[28])*sys_part_1;
  grad_3_num += 6*(60*c1p1*c3p1*c5p1*c7p2*ss[30] + 60*c1p1*c3p1*c5p2*c7p1*ss[28] + 60*c1p1*c3p2*c5p1*c7p1*ss[26] + 60*c1p2*c3p1*c5p1*c7p1*ss[24] + 20*c1p1*c3p1*c5p3*ss[26] + 20*c1p1*c3p1*c7p3*ss[32] + 30*c1p1*c3p2*c5p2*ss[24] + 30*c1p1*c3p2*c7p2*ss[28] + 20*c1p1*c3p3*c5p1*ss[22] + 20*c1p1*c3p3*c7p1*ss[24] + 20*c1p1*c5p1*c7p3*ss[34] + 30*c1p1*c5p2*c7p2*ss[32] + 20*c1p1*c5p3*c7p1*ss[30] + 30*c1p2*c3p1*c5p2*ss[22] + 30*c1p2*c3p1*c7p2*ss[26] + 30*c1p2*c3p2*c5p1*ss[20] + 30*c1p2*c3p2*c7p1*ss[22] + 30*c1p2*c5p1*c7p2*ss[28] + 30*c1p2*c5p2*c7p1*ss[26] + 20*c1p3*c3p1*c5p1*ss[18] + 20*c1p3*c3p1*c7p1*ss[20] + 20*c1p3*c5p1*c7p1*ss[22] + 20*c3p1*c5p1*c7p3*ss[36] + 30*c3p1*c5p2*c7p2*ss[34] + 20*c3p1*c5p3*c7p1*ss[32] + 30*c3p2*c5p1*c7p2*ss[32] + 30*c3p2*c5p2*c7p1*ss[30] + 20*c3p3*c5p1*c7p1*ss[28] + 5*c1p1*c3p4*ss[20] + 5*c1p1*c5p4*ss[28] + 5*c1p1*c7p4*ss[36] + 10*c1p2*c3p3*ss[18] + 10*c1p2*c5p3*ss[24] + 10*c1p2*c7p3*ss[30] + 10*c1p3*c3p2*ss[16] + 10*c1p3*c5p2*ss[20] + 10*c1p3*c7p2*ss[24] + 5*c1p4*c3p1*ss[14] + 5*c1p4*c5p1*ss[16] + 5*c1p4*c7p1*ss[18] + 5*c3p1*c5p4*ss[30] + 5*c3p1*c7p4*ss[38] + 10*c3p2*c5p3*ss[28] + 10*c3p2*c7p3*ss[34] + 10*c3p3*c5p2*ss[26] + 10*c3p3*c7p2*ss[30] + 5*c3p4*c5p1*ss[24] + 5*c3p4*c7p1*ss[26] + 5*c5p1*c7p4*ss[40] + 10*c5p2*c7p3*ss[38] + 10*c5p3*c7p2*ss[36] + 5*c5p4*c7p1*ss[34] + c1p5*ss[12] + c3p5*ss[22] + c5p5*ss[32] + c7p5*ss[42])*sys_part_2;
  grad_3_num += 8*(210*c1p1*c3p1*c5p1*c7p4*ss[44] + 420*c1p1*c3p1*c5p2*c7p3*ss[42] + 420*c1p1*c3p1*c5p3*c7p2*ss[40] + 210*c1p1*c3p1*c5p4*c7p1*ss[38] + 420*c1p1*c3p2*c5p1*c7p3*ss[40] + 630*c1p1*c3p2*c5p2*c7p2*ss[38] + 420*c1p1*c3p2*c5p3*c7p1*ss[36] + 420*c1p1*c3p3*c5p1*c7p2*ss[36] + 420*c1p1*c3p3*c5p2*c7p1*ss[34] + 210*c1p1*c3p4*c5p1*c7p1*ss[32] + 420*c1p2*c3p1*c5p1*c7p3*ss[38] + 630*c1p2*c3p1*c5p2*c7p2*ss[36] + 420*c1p2*c3p1*c5p3*c7p1*ss[34] + 630*c1p2*c3p2*c5p1*c7p2*ss[34] + 630*c1p2*c3p2*c5p2*c7p1*ss[32] + 420*c1p2*c3p3*c5p1*c7p1*ss[30] + 420*c1p3*c3p1*c5p1*c7p2*ss[32] + 420*c1p3*c3p1*c5p2*c7p1*ss[30] + 420*c1p3*c3p2*c5p1*c7p1*ss[28] + 210*c1p4*c3p1*c5p1*c7p1*ss[26] + 42*c1p1*c3p1*c5p5*ss[36] + 42*c1p1*c3p1*c7p5*ss[46] + 105*c1p1*c3p2*c5p4*ss[34] + 105*c1p1*c3p2*c7p4*ss[42] + 140*c1p1*c3p3*c5p3*ss[32] + 140*c1p1*c3p3*c7p3*ss[38] + 105*c1p1*c3p4*c5p2*ss[30] + 105*c1p1*c3p4*c7p2*ss[34] + 42*c1p1*c3p5*c5p1*ss[28] + 42*c1p1*c3p5*c7p1*ss[30] + 42*c1p1*c5p1*c7p5*ss[48] + 105*c1p1*c5p2*c7p4*ss[46] + 140*c1p1*c5p3*c7p3*ss[44] + 105*c1p1*c5p4*c7p2*ss[42] + 42*c1p1*c5p5*c7p1*ss[40] + 105*c1p2*c3p1*c5p4*ss[32] + 105*c1p2*c3p1*c7p4*ss[40] + 210*c1p2*c3p2*c5p3*ss[30] + 210*c1p2*c3p2*c7p3*ss[36] + 210*c1p2*c3p3*c5p2*ss[28] + 210*c1p2*c3p3*c7p2*ss[32] + 105*c1p2*c3p4*c5p1*ss[26] + 105*c1p2*c3p4*c7p1*ss[28] + 105*c1p2*c5p1*c7p4*ss[42] + 210*c1p2*c5p2*c7p3*ss[40] + 210*c1p2*c5p3*c7p2*ss[38] + 105*c1p2*c5p4*c7p1*ss[36] + 140*c1p3*c3p1*c5p3*ss[28] + 140*c1p3*c3p1*c7p3*ss[34] + 210*c1p3*c3p2*c5p2*ss[26] + 210*c1p3*c3p2*c7p2*ss[30] + 140*c1p3*c3p3*c5p1*ss[24] + 140*c1p3*c3p3*c7p1*ss[26] + 140*c1p3*c5p1*c7p3*ss[36] + 210*c1p3*c5p2*c7p2*ss[34] + 140*c1p3*c5p3*c7p1*ss[32] + 105*c1p4*c3p1*c5p2*ss[24] + 105*c1p4*c3p1*c7p2*ss[28] + 105*c1p4*c3p2*c5p1*ss[22] + 105*c1p4*c3p2*c7p1*ss[24] + 105*c1p4*c5p1*c7p2*ss[30] + 105*c1p4*c5p2*c7p1*ss[28] + 42*c1p5*c3p1*c5p1*ss[20] + 42*c1p5*c3p1*c7p1*ss[22] + 42*c1p5*c5p1*c7p1*ss[24] + 42*c3p1*c5p1*c7p5*ss[50] + 105*c3p1*c5p2*c7p4*ss[48] + 140*c3p1*c5p3*c7p3*ss[46] + 105*c3p1*c5p4*c7p2*ss[44] + 42*c3p1*c5p5*c7p1*ss[42] + 105*c3p2*c5p1*c7p4*ss[46] + 210*c3p2*c5p2*c7p3*ss[44] + 210*c3p2*c5p3*c7p2*ss[42] + 105*c3p2*c5p4*c7p1*ss[40] + 140*c3p3*c5p1*c7p3*ss[42] + 210*c3p3*c5p2*c7p2*ss[40] + 140*c3p3*c5p3*c7p1*ss[38] + 105*c3p4*c5p1*c7p2*ss[38] + 105*c3p4*c5p2*c7p1*ss[36] + 42*c3p5*c5p1*c7p1*ss[34] + 7*c1p1*c3p6*ss[26] + 7*c1p1*c5p6*ss[38] + 7*c1p1*c7p6*ss[50] + 21*c1p2*c3p5*ss[24] + 21*c1p2*c5p5*ss[34] + 21*c1p2*c7p5*ss[44] + 35*c1p3*c3p4*ss[22] + 35*c1p3*c5p4*ss[30] + 35*c1p3*c7p4*ss[38] + 35*c1p4*c3p3*ss[20] + 35*c1p4*c5p3*ss[26] + 35*c1p4*c7p3*ss[32] + 21*c1p5*c3p2*ss[18] + 21*c1p5*c5p2*ss[22] + 21*c1p5*c7p2*ss[26] + 7*c1p6*c3p1*ss[16] + 7*c1p6*c5p1*ss[18] + 7*c1p6*c7p1*ss[20] + 7*c3p1*c5p6*ss[40] + 7*c3p1*c7p6*ss[52] + 21*c3p2*c5p5*ss[38] + 21*c3p2*c7p5*ss[48] + 35*c3p3*c5p4*ss[36] + 35*c3p3*c7p4*ss[44] + 35*c3p4*c5p3*ss[34] + 35*c3p4*c7p3*ss[40] + 21*c3p5*c5p2*ss[32] + 21*c3p5*c7p2*ss[36] + 7*c3p6*c5p1*ss[30] + 7*c3p6*c7p1*ss[32] + 7*c5p1*c7p6*ss[54] + 21*c5p2*c7p5*ss[52] + 35*c5p3*c7p4*ss[50] + 35*c5p4*c7p3*ss[48] + 21*c5p5*c7p2*ss[46] + 7*c5p6*c7p1*ss[44] + c1p7*ss[14] + c3p7*ss[28] + c5p7*ss[42] + c7p7*ss[56])*sys_part_3;
  TACC grad_3_den = sqrt(square(sys_part_0) + square(sys_part_1) + square(sys_part_2) + square(sys_part_3));
  TACC metric_sqr = convert<TACC>(square(sys_part_0));
  metric_sqr += square(sys_part_1);
  metric_sqr += square(sys_part_2);
  metric_sqr += square(sys_part_3);
  gradient[0]=convert<T>(grad_0_num / grad_0_den);
  gradient[1]=convert<T>(grad_1_num / grad_1_den);
  gradient[2]=convert<T>(grad_2_num / grad_2_den);
  gradient[3]=convert<T>(grad_3_num / grad_3_den);
  return convert<T>(sqrt(metric_sqr));
}
void roots(const T *parameters, T *roots) const {
  T c1p1 = parameters[0];
  T c3p1 = parameters[1];
  T c5p1 = parameters[2];
  T c7p1 = parameters[3];
  T c1p2 = c1p1*c1p1;
  T c3p2 = c3p1*c3p1;
  T c5p2 = c5p1*c5p1;
  T c7p2 = c7p1*c7p1;
  TACC root_0 = convert<TACC>(2*c1p1*c3p1*ss[4]);
  root_0 += 2*c1p1*c5p1*ss[6];
  root_0 += 2*c1p1*c7p1*ss[8];
  root_0 += 2*c3p1*c5p1*ss[8];
  root_0 += 2*c3p1*c7p1*ss[10];
  root_0 += 2*c5p1*c7p1*ss[12];
  root_0 += c1p2*ss[2];
  root_0 += c3p2*ss[6];
  root_0 += c5p2*ss[10];
  root_0 += c7p2*ss[14];
  root_0 += -tmom[2];
  T c1p3 = c1p2*c1p1;
  T c1p4 = c1p3*c1p1;
  T c3p3 = c3p2*c3p1;
  T c3p4 = c3p3*c3p1;
  T c5p3 = c5p2*c5p1;
  T c5p4 = c5p3*c5p1;
  T c7p3 = c7p2*c7p1;
  T c7p4 = c7p3*c7p1;
  TACC root_1 = convert<TACC>(24*c1p1*c3p1*c5p1*c7p1*ss[16]);
  root_1 += 12*c1p1*c3p1*c5p2*ss[14];
  root_1 += 12*c1p1*c3p1*c7p2*ss[18];
  root_1 += 12*c1p1*c3p2*c5p1*ss[12];
  root_1 += 12*c1p1*c3p2*c7p1*ss[14];
  root_1 += 12*c1p1*c5p1*c7p2*ss[20];
  root_1 += 12*c1p1*c5p2*c7p1*ss[18];
  root_1 += 12*c1p2*c3p1*c5p1*ss[10];
  root_1 += 12*c1p2*c3p1*c7p1*ss[12];
  root_1 += 12*c1p2*c5p1*c7p1*ss[14];
  root_1 += 12*c3p1*c5p1*c7p2*ss[22];
  root_1 += 12*c3p1*c5p2*c7p1*ss[20];
  root_1 += 12*c3p2*c5p1*c7p1*ss[18];
  root_1 += 4*c1p1*c3p3*ss[10];
  root_1 += 4*c1p1*c5p3*ss[16];
  root_1 += 4*c1p1*c7p3*ss[22];
  root_1 += 6*c1p2*c3p2*ss[8];
  root_1 += 6*c1p2*c5p2*ss[12];
  root_1 += 6*c1p2*c7p2*ss[16];
  root_1 += 4*c1p3*c3p1*ss[6];
  root_1 += 4*c1p3*c5p1*ss[8];
  root_1 += 4*c1p3*c7p1*ss[10];
  root_1 += 4*c3p1*c5p3*ss[18];
  root_1 += 4*c3p1*c7p3*ss[24];
  root_1 += 6*c3p2*c5p2*ss[16];
  root_1 += 6*c3p2*c7p2*ss[20];
  root_1 += 4*c3p3*c5p1*ss[14];
  root_1 += 4*c3p3*c7p1*ss[16];
  root_1 += 4*c5p1*c7p3*ss[26];
  root_1 += 6*c5p2*c7p2*ss[24];
  root_1 += 4*c5p3*c7p1*ss[22];
  root_1 += c1p4*ss[4];
  root_1 += c3p4*ss[12];
  root_1 += c5p4*ss[20];
  root_1 += c7p4*ss[28];
  root_1 += -tmom[4];
  T c1p5 = c1p4*c1p1;
  T c1p6 = c1p5*c1p1;
  T c3p5 = c3p4*c3p1;
  T c3p6 = c3p5*c3p1;
  T c5p5 = c5p4*c5p1;
  T c5p6 = c5p5*c5p1;
  T c7p5 = c7p4*c7p1;
  T c7p6 = c7p5*c7p1;
  TACC root_2 = convert<TACC>(120*c1p1*c3p1*c5p1*c7p3*ss[30]);
  root_2 += 180*c1p1*c3p1*c5p2*c7p2*ss[28];
  root_2 += 120*c1p1*c3p1*c5p3*c7p1*ss[26];
  root_2 += 180*c1p1*c3p2*c5p1*c7p2*ss[26];
  root_2 += 180*c1p1*c3p2*c5p2*c7p1*ss[24];
  root_2 += 120*c1p1*c3p3*c5p1*c7p1*ss[22];
  root_2 += 180*c1p2*c3p1*c5p1*c7p2*ss[24];
  root_2 += 180*c1p2*c3p1*c5p2*c7p1*ss[22];
  root_2 += 180*c1p2*c3p2*c5p1*c7p1*ss[20];
  root_2 += 120*c1p3*c3p1*c5p1*c7p1*ss[18];
  root_2 += 30*c1p1*c3p1*c5p4*ss[24];
  root_2 += 30*c1p1*c3p1*c7p4*ss[32];
  root_2 += 60*c1p1*c3p2*c5p3*ss[22];
  root_2 += 60*c1p1*c3p2*c7p3*ss[28];
  root_2 += 60*c1p1*c3p3*c5p2*ss[20];
  root_2 += 60*c1p1*c3p3*c7p2*ss[24];
  root_2 += 30*c1p1*c3p4*c5p1*ss[18];
  root_2 += 30*c1p1*c3p4*c7p1*ss[20];
  root_2 += 30*c1p1*c5p1*c7p4*ss[34];
  root_2 += 60*c1p1*c5p2*c7p3*ss[32];
  root_2 += 60*c1p1*c5p3*c7p2*ss[30];
  root_2 += 30*c1p1*c5p4*c7p1*ss[28];
  root_2 += 60*c1p2*c3p1*c5p3*ss[20];
  root_2 += 60*c1p2*c3p1*c7p3*ss[26];
  root_2 += 90*c1p2*c3p2*c5p2*ss[18];
  root_2 += 90*c1p2*c3p2*c7p2*ss[22];
  root_2 += 60*c1p2*c3p3*c5p1*ss[16];
  root_2 += 60*c1p2*c3p3*c7p1*ss[18];
  root_2 += 60*c1p2*c5p1*c7p3*ss[28];
  root_2 += 90*c1p2*c5p2*c7p2*ss[26];
  root_2 += 60*c1p2*c5p3*c7p1*ss[24];
  root_2 += 60*c1p3*c3p1*c5p2*ss[16];
  root_2 += 60*c1p3*c3p1*c7p2*ss[20];
  root_2 += 60*c1p3*c3p2*c5p1*ss[14];
  root_2 += 60*c1p3*c3p2*c7p1*ss[16];
  root_2 += 60*c1p3*c5p1*c7p2*ss[22];
  root_2 += 60*c1p3*c5p2*c7p1*ss[20];
  root_2 += 30*c1p4*c3p1*c5p1*ss[12];
  root_2 += 30*c1p4*c3p1*c7p1*ss[14];
  root_2 += 30*c1p4*c5p1*c7p1*ss[16];
  root_2 += 30*c3p1*c5p1*c7p4*ss[36];
  root_2 += 60*c3p1*c5p2*c7p3*ss[34];
  root_2 += 60*c3p1*c5p3*c7p2*ss[32];
  root_2 += 30*c3p1*c5p4*c7p1*ss[30];
  root_2 += 60*c3p2*c5p1*c7p3*ss[32];
  root_2 += 90*c3p2*c5p2*c7p2*ss[30];
  root_2 += 60*c3p2*c5p3*c7p1*ss[28];
  root_2 += 60*c3p3*c5p1*c7p2*ss[28];
  root_2 += 60*c3p3*c5p2*c7p1*ss[26];
  root_2 += 30*c3p4*c5p1*c7p1*ss[24];
  root_2 += 6*c1p1*c3p5*ss[16];
  root_2 += 6*c1p1*c5p5*ss[26];
  root_2 += 6*c1p1*c7p5*ss[36];
  root_2 += 15*c1p2*c3p4*ss[14];
  root_2 += 15*c1p2*c5p4*ss[22];
  root_2 += 15*c1p2*c7p4*ss[30];
  root_2 += 20*c1p3*c3p3*ss[12];
  root_2 += 20*c1p3*c5p3*ss[18];
  root_2 += 20*c1p3*c7p3*ss[24];
  root_2 += 15*c1p4*c3p2*ss[10];
  root_2 += 15*c1p4*c5p2*ss[14];
  root_2 += 15*c1p4*c7p2*ss[18];
  root_2 += 6*c1p5*c3p1*ss[8];
  root_2 += 6*c1p5*c5p1*ss[10];
  root_2 += 6*c1p5*c7p1*ss[12];
  root_2 += 6*c3p1*c5p5*ss[28];
  root_2 += 6*c3p1*c7p5*ss[38];
  root_2 += 15*c3p2*c5p4*ss[26];
  root_2 += 15*c3p2*c7p4*ss[34];
  root_2 += 20*c3p3*c5p3*ss[24];
  root_2 += 20*c3p3*c7p3*ss[30];
  root_2 += 15*c3p4*c5p2*ss[22];
  root_2 += 15*c3p4*c7p2*ss[26];
  root_2 += 6*c3p5*c5p1*ss[20];
  root_2 += 6*c3p5*c7p1*ss[22];
  root_2 += 6*c5p1*c7p5*ss[40];
  root_2 += 15*c5p2*c7p4*ss[38];
  root_2 += 20*c5p3*c7p3*ss[36];
  root_2 += 15*c5p4*c7p2*ss[34];
  root_2 += 6*c5p5*c7p1*ss[32];
  root_2 += c1p6*ss[6];
  root_2 += c3p6*ss[18];
  root_2 += c5p6*ss[30];
  root_2 += c7p6*ss[42];
  root_2 += -tmom[6];
  T c1p7 = c1p6*c1p1;
  T c1p8 = c1p7*c1p1;
  T c3p7 = c3p6*c3p1;
  T c3p8 = c3p7*c3p1;
  T c5p7 = c5p6*c5p1;
  T c5p8 = c5p7*c5p1;
  T c7p7 = c7p6*c7p1;
  T c7p8 = c7p7*c7p1;
  TACC root_3 = convert<TACC>(336*c1p1*c3p1*c5p1*c7p5*ss[44]);
  root_3 += 840*c1p1*c3p1*c5p2*c7p4*ss[42];
  root_3 += 1120*c1p1*c3p1*c5p3*c7p3*ss[40];
  root_3 += 840*c1p1*c3p1*c5p4*c7p2*ss[38];
  root_3 += 336*c1p1*c3p1*c5p5*c7p1*ss[36];
  root_3 += 840*c1p1*c3p2*c5p1*c7p4*ss[40];
  root_3 += 1680*c1p1*c3p2*c5p2*c7p3*ss[38];
  root_3 += 1680*c1p1*c3p2*c5p3*c7p2*ss[36];
  root_3 += 840*c1p1*c3p2*c5p4*c7p1*ss[34];
  root_3 += 1120*c1p1*c3p3*c5p1*c7p3*ss[36];
  root_3 += 1680*c1p1*c3p3*c5p2*c7p2*ss[34];
  root_3 += 1120*c1p1*c3p3*c5p3*c7p1*ss[32];
  root_3 += 840*c1p1*c3p4*c5p1*c7p2*ss[32];
  root_3 += 840*c1p1*c3p4*c5p2*c7p1*ss[30];
  root_3 += 336*c1p1*c3p5*c5p1*c7p1*ss[28];
  root_3 += 840*c1p2*c3p1*c5p1*c7p4*ss[38];
  root_3 += 1680*c1p2*c3p1*c5p2*c7p3*ss[36];
  root_3 += 1680*c1p2*c3p1*c5p3*c7p2*ss[34];
  root_3 += 840*c1p2*c3p1*c5p4*c7p1*ss[32];
  root_3 += 1680*c1p2*c3p2*c5p1*c7p3*ss[34];
  root_3 += 2520*c1p2*c3p2*c5p2*c7p2*ss[32];
  root_3 += 1680*c1p2*c3p2*c5p3*c7p1*ss[30];
  root_3 += 1680*c1p2*c3p3*c5p1*c7p2*ss[30];
  root_3 += 1680*c1p2*c3p3*c5p2*c7p1*ss[28];
  root_3 += 840*c1p2*c3p4*c5p1*c7p1*ss[26];
  root_3 += 1120*c1p3*c3p1*c5p1*c7p3*ss[32];
  root_3 += 1680*c1p3*c3p1*c5p2*c7p2*ss[30];
  root_3 += 1120*c1p3*c3p1*c5p3*c7p1*ss[28];
  root_3 += 1680*c1p3*c3p2*c5p1*c7p2*ss[28];
  root_3 += 1680*c1p3*c3p2*c5p2*c7p1*ss[26];
  root_3 += 1120*c1p3*c3p3*c5p1*c7p1*ss[24];
  root_3 += 840*c1p4*c3p1*c5p1*c7p2*ss[26];
  root_3 += 840*c1p4*c3p1*c5p2*c7p1*ss[24];
  root_3 += 840*c1p4*c3p2*c5p1*c7p1*ss[22];
  root_3 += 336*c1p5*c3p1*c5p1*c7p1*ss[20];
  root_3 += 56*c1p1*c3p1*c5p6*ss[34];
  root_3 += 56*c1p1*c3p1*c7p6*ss[46];
  root_3 += 168*c1p1*c3p2*c5p5*ss[32];
  root_3 += 168*c1p1*c3p2*c7p5*ss[42];
  root_3 += 280*c1p1*c3p3*c5p4*ss[30];
  root_3 += 280*c1p1*c3p3*c7p4*ss[38];
  root_3 += 280*c1p1*c3p4*c5p3*ss[28];
  root_3 += 280*c1p1*c3p4*c7p3*ss[34];
  root_3 += 168*c1p1*c3p5*c5p2*ss[26];
  root_3 += 168*c1p1*c3p5*c7p2*ss[30];
  root_3 += 56*c1p1*c3p6*c5p1*ss[24];
  root_3 += 56*c1p1*c3p6*c7p1*ss[26];
  root_3 += 56*c1p1*c5p1*c7p6*ss[48];
  root_3 += 168*c1p1*c5p2*c7p5*ss[46];
  root_3 += 280*c1p1*c5p3*c7p4*ss[44];
  root_3 += 280*c1p1*c5p4*c7p3*ss[42];
  root_3 += 168*c1p1*c5p5*c7p2*ss[40];
  root_3 += 56*c1p1*c5p6*c7p1*ss[38];
  root_3 += 168*c1p2*c3p1*c5p5*ss[30];
  root_3 += 168*c1p2*c3p1*c7p5*ss[40];
  root_3 += 420*c1p2*c3p2*c5p4*ss[28];
  root_3 += 420*c1p2*c3p2*c7p4*ss[36];
  root_3 += 560*c1p2*c3p3*c5p3*ss[26];
  root_3 += 560*c1p2*c3p3*c7p3*ss[32];
  root_3 += 420*c1p2*c3p4*c5p2*ss[24];
  root_3 += 420*c1p2*c3p4*c7p2*ss[28];
  root_3 += 168*c1p2*c3p5*c5p1*ss[22];
  root_3 += 168*c1p2*c3p5*c7p1*ss[24];
  root_3 += 168*c1p2*c5p1*c7p5*ss[42];
  root_3 += 420*c1p2*c5p2*c7p4*ss[40];
  root_3 += 560*c1p2*c5p3*c7p3*ss[38];
  root_3 += 420*c1p2*c5p4*c7p2*ss[36];
  root_3 += 168*c1p2*c5p5*c7p1*ss[34];
  root_3 += 280*c1p3*c3p1*c5p4*ss[26];
  root_3 += 280*c1p3*c3p1*c7p4*ss[34];
  root_3 += 560*c1p3*c3p2*c5p3*ss[24];
  root_3 += 560*c1p3*c3p2*c7p3*ss[30];
  root_3 += 560*c1p3*c3p3*c5p2*ss[22];
  root_3 += 560*c1p3*c3p3*c7p2*ss[26];
  root_3 += 280*c1p3*c3p4*c5p1*ss[20];
  root_3 += 280*c1p3*c3p4*c7p1*ss[22];
  root_3 += 280*c1p3*c5p1*c7p4*ss[36];
  root_3 += 560*c1p3*c5p2*c7p3*ss[34];
  root_3 += 560*c1p3*c5p3*c7p2*ss[32];
  root_3 += 280*c1p3*c5p4*c7p1*ss[30];
  root_3 += 280*c1p4*c3p1*c5p3*ss[22];
  root_3 += 280*c1p4*c3p1*c7p3*ss[28];
  root_3 += 420*c1p4*c3p2*c5p2*ss[20];
  root_3 += 420*c1p4*c3p2*c7p2*ss[24];
  root_3 += 280*c1p4*c3p3*c5p1*ss[18];
  root_3 += 280*c1p4*c3p3*c7p1*ss[20];
  root_3 += 280*c1p4*c5p1*c7p3*ss[30];
  root_3 += 420*c1p4*c5p2*c7p2*ss[28];
  root_3 += 280*c1p4*c5p3*c7p1*ss[26];
  root_3 += 168*c1p5*c3p1*c5p2*ss[18];
  root_3 += 168*c1p5*c3p1*c7p2*ss[22];
  root_3 += 168*c1p5*c3p2*c5p1*ss[16];
  root_3 += 168*c1p5*c3p2*c7p1*ss[18];
  root_3 += 168*c1p5*c5p1*c7p2*ss[24];
  root_3 += 168*c1p5*c5p2*c7p1*ss[22];
  root_3 += 56*c1p6*c3p1*c5p1*ss[14];
  root_3 += 56*c1p6*c3p1*c7p1*ss[16];
  root_3 += 56*c1p6*c5p1*c7p1*ss[18];
  root_3 += 56*c3p1*c5p1*c7p6*ss[50];
  root_3 += 168*c3p1*c5p2*c7p5*ss[48];
  root_3 += 280*c3p1*c5p3*c7p4*ss[46];
  root_3 += 280*c3p1*c5p4*c7p3*ss[44];
  root_3 += 168*c3p1*c5p5*c7p2*ss[42];
  root_3 += 56*c3p1*c5p6*c7p1*ss[40];
  root_3 += 168*c3p2*c5p1*c7p5*ss[46];
  root_3 += 420*c3p2*c5p2*c7p4*ss[44];
  root_3 += 560*c3p2*c5p3*c7p3*ss[42];
  root_3 += 420*c3p2*c5p4*c7p2*ss[40];
  root_3 += 168*c3p2*c5p5*c7p1*ss[38];
  root_3 += 280*c3p3*c5p1*c7p4*ss[42];
  root_3 += 560*c3p3*c5p2*c7p3*ss[40];
  root_3 += 560*c3p3*c5p3*c7p2*ss[38];
  root_3 += 280*c3p3*c5p4*c7p1*ss[36];
  root_3 += 280*c3p4*c5p1*c7p3*ss[38];
  root_3 += 420*c3p4*c5p2*c7p2*ss[36];
  root_3 += 280*c3p4*c5p3*c7p1*ss[34];
  root_3 += 168*c3p5*c5p1*c7p2*ss[34];
  root_3 += 168*c3p5*c5p2*c7p1*ss[32];
  root_3 += 56*c3p6*c5p1*c7p1*ss[30];
  root_3 += 8*c1p1*c3p7*ss[22];
  root_3 += 8*c1p1*c5p7*ss[36];
  root_3 += 8*c1p1*c7p7*ss[50];
  root_3 += 28*c1p2*c3p6*ss[20];
  root_3 += 28*c1p2*c5p6*ss[32];
  root_3 += 28*c1p2*c7p6*ss[44];
  root_3 += 56*c1p3*c3p5*ss[18];
  root_3 += 56*c1p3*c5p5*ss[28];
  root_3 += 56*c1p3*c7p5*ss[38];
  root_3 += 70*c1p4*c3p4*ss[16];
  root_3 += 70*c1p4*c5p4*ss[24];
  root_3 += 70*c1p4*c7p4*ss[32];
  root_3 += 56*c1p5*c3p3*ss[14];
  root_3 += 56*c1p5*c5p3*ss[20];
  root_3 += 56*c1p5*c7p3*ss[26];
  root_3 += 28*c1p6*c3p2*ss[12];
  root_3 += 28*c1p6*c5p2*ss[16];
  root_3 += 28*c1p6*c7p2*ss[20];
  root_3 += 8*c1p7*c3p1*ss[10];
  root_3 += 8*c1p7*c5p1*ss[12];
  root_3 += 8*c1p7*c7p1*ss[14];
  root_3 += 8*c3p1*c5p7*ss[38];
  root_3 += 8*c3p1*c7p7*ss[52];
  root_3 += 28*c3p2*c5p6*ss[36];
  root_3 += 28*c3p2*c7p6*ss[48];
  root_3 += 56*c3p3*c5p5*ss[34];
  root_3 += 56*c3p3*c7p5*ss[44];
  root_3 += 70*c3p4*c5p4*ss[32];
  root_3 += 70*c3p4*c7p4*ss[40];
  root_3 += 56*c3p5*c5p3*ss[30];
  root_3 += 56*c3p5*c7p3*ss[36];
  root_3 += 28*c3p6*c5p2*ss[28];
  root_3 += 28*c3p6*c7p2*ss[32];
  root_3 += 8*c3p7*c5p1*ss[26];
  root_3 += 8*c3p7*c7p1*ss[28];
  root_3 += 8*c5p1*c7p7*ss[54];
  root_3 += 28*c5p2*c7p6*ss[52];
  root_3 += 56*c5p3*c7p5*ss[50];
  root_3 += 70*c5p4*c7p4*ss[48];
  root_3 += 56*c5p5*c7p3*ss[46];
  root_3 += 28*c5p6*c7p2*ss[44];
  root_3 += 8*c5p7*c7p1*ss[42];
  root_3 += c1p8*ss[8];
  root_3 += c3p8*ss[24];
  root_3 += c5p8*ss[40];
  root_3 += c7p8*ss[56];
  root_3 += -tmom[8];
  roots[0]=convert<T>(root_0);
  roots[1]=convert<T>(root_1);
  roots[2]=convert<T>(root_2);
  roots[3]=convert<T>(root_3);
}
void roots_and_jacobian(const T *parameters, T *roots, T *jacobian) const {
  T c1p1 = parameters[0];
  T c3p1 = parameters[1];
  T c5p1 = parameters[2];
  T c7p1 = parameters[3];
  T c1p2 = c1p1*c1p1;
  T c3p2 = c3p1*c3p1;
  T c5p2 = c5p1*c5p1;
  T c7p2 = c7p1*c7p1;
  TACC root_0 = convert<TACC>(2*c1p1*c3p1*ss[4]);
  root_0 += 2*c1p1*c5p1*ss[6];
  root_0 += 2*c1p1*c7p1*ss[8];
  root_0 += 2*c3p1*c5p1*ss[8];
  root_0 += 2*c3p1*c7p1*ss[10];
  root_0 += 2*c5p1*c7p1*ss[12];
  root_0 += c1p2*ss[2];
  root_0 += c3p2*ss[6];
  root_0 += c5p2*ss[10];
  root_0 += c7p2*ss[14];
  root_0 += -tmom[2];
  T c1p3 = c1p2*c1p1;
  T c1p4 = c1p3*c1p1;
  T c3p3 = c3p2*c3p1;
  T c3p4 = c3p3*c3p1;
  T c5p3 = c5p2*c5p1;
  T c5p4 = c5p3*c5p1;
  T c7p3 = c7p2*c7p1;
  T c7p4 = c7p3*c7p1;
  TACC root_1 = convert<TACC>(24*c1p1*c3p1*c5p1*c7p1*ss[16]);
  root_1 += 12*c1p1*c3p1*c5p2*ss[14];
  root_1 += 12*c1p1*c3p1*c7p2*ss[18];
  root_1 += 12*c1p1*c3p2*c5p1*ss[12];
  root_1 += 12*c1p1*c3p2*c7p1*ss[14];
  root_1 += 12*c1p1*c5p1*c7p2*ss[20];
  root_1 += 12*c1p1*c5p2*c7p1*ss[18];
  root_1 += 12*c1p2*c3p1*c5p1*ss[10];
  root_1 += 12*c1p2*c3p1*c7p1*ss[12];
  root_1 += 12*c1p2*c5p1*c7p1*ss[14];
  root_1 += 12*c3p1*c5p1*c7p2*ss[22];
  root_1 += 12*c3p1*c5p2*c7p1*ss[20];
  root_1 += 12*c3p2*c5p1*c7p1*ss[18];
  root_1 += 4*c1p1*c3p3*ss[10];
  root_1 += 4*c1p1*c5p3*ss[16];
  root_1 += 4*c1p1*c7p3*ss[22];
  root_1 += 6*c1p2*c3p2*ss[8];
  root_1 += 6*c1p2*c5p2*ss[12];
  root_1 += 6*c1p2*c7p2*ss[16];
  root_1 += 4*c1p3*c3p1*ss[6];
  root_1 += 4*c1p3*c5p1*ss[8];
  root_1 += 4*c1p3*c7p1*ss[10];
  root_1 += 4*c3p1*c5p3*ss[18];
  root_1 += 4*c3p1*c7p3*ss[24];
  root_1 += 6*c3p2*c5p2*ss[16];
  root_1 += 6*c3p2*c7p2*ss[20];
  root_1 += 4*c3p3*c5p1*ss[14];
  root_1 += 4*c3p3*c7p1*ss[16];
  root_1 += 4*c5p1*c7p3*ss[26];
  root_1 += 6*c5p2*c7p2*ss[24];
  root_1 += 4*c5p3*c7p1*ss[22];
  root_1 += c1p4*ss[4];
  root_1 += c3p4*ss[12];
  root_1 += c5p4*ss[20];
  root_1 += c7p4*ss[28];
  root_1 += -tmom[4];
  T c1p5 = c1p4*c1p1;
  T c1p6 = c1p5*c1p1;
  T c3p5 = c3p4*c3p1;
  T c3p6 = c3p5*c3p1;
  T c5p5 = c5p4*c5p1;
  T c5p6 = c5p5*c5p1;
  T c7p5 = c7p4*c7p1;
  T c7p6 = c7p5*c7p1;
  TACC root_2 = convert<TACC>(120*c1p1*c3p1*c5p1*c7p3*ss[30]);
  root_2 += 180*c1p1*c3p1*c5p2*c7p2*ss[28];
  root_2 += 120*c1p1*c3p1*c5p3*c7p1*ss[26];
  root_2 += 180*c1p1*c3p2*c5p1*c7p2*ss[26];
  root_2 += 180*c1p1*c3p2*c5p2*c7p1*ss[24];
  root_2 += 120*c1p1*c3p3*c5p1*c7p1*ss[22];
  root_2 += 180*c1p2*c3p1*c5p1*c7p2*ss[24];
  root_2 += 180*c1p2*c3p1*c5p2*c7p1*ss[22];
  root_2 += 180*c1p2*c3p2*c5p1*c7p1*ss[20];
  root_2 += 120*c1p3*c3p1*c5p1*c7p1*ss[18];
  root_2 += 30*c1p1*c3p1*c5p4*ss[24];
  root_2 += 30*c1p1*c3p1*c7p4*ss[32];
  root_2 += 60*c1p1*c3p2*c5p3*ss[22];
  root_2 += 60*c1p1*c3p2*c7p3*ss[28];
  root_2 += 60*c1p1*c3p3*c5p2*ss[20];
  root_2 += 60*c1p1*c3p3*c7p2*ss[24];
  root_2 += 30*c1p1*c3p4*c5p1*ss[18];
  root_2 += 30*c1p1*c3p4*c7p1*ss[20];
  root_2 += 30*c1p1*c5p1*c7p4*ss[34];
  root_2 += 60*c1p1*c5p2*c7p3*ss[32];
  root_2 += 60*c1p1*c5p3*c7p2*ss[30];
  root_2 += 30*c1p1*c5p4*c7p1*ss[28];
  root_2 += 60*c1p2*c3p1*c5p3*ss[20];
  root_2 += 60*c1p2*c3p1*c7p3*ss[26];
  root_2 += 90*c1p2*c3p2*c5p2*ss[18];
  root_2 += 90*c1p2*c3p2*c7p2*ss[22];
  root_2 += 60*c1p2*c3p3*c5p1*ss[16];
  root_2 += 60*c1p2*c3p3*c7p1*ss[18];
  root_2 += 60*c1p2*c5p1*c7p3*ss[28];
  root_2 += 90*c1p2*c5p2*c7p2*ss[26];
  root_2 += 60*c1p2*c5p3*c7p1*ss[24];
  root_2 += 60*c1p3*c3p1*c5p2*ss[16];
  root_2 += 60*c1p3*c3p1*c7p2*ss[20];
  root_2 += 60*c1p3*c3p2*c5p1*ss[14];
  root_2 += 60*c1p3*c3p2*c7p1*ss[16];
  root_2 += 60*c1p3*c5p1*c7p2*ss[22];
  root_2 += 60*c1p3*c5p2*c7p1*ss[20];
  root_2 += 30*c1p4*c3p1*c5p1*ss[12];
  root_2 += 30*c1p4*c3p1*c7p1*ss[14];
  root_2 += 30*c1p4*c5p1*c7p1*ss[16];
  root_2 += 30*c3p1*c5p1*c7p4*ss[36];
  root_2 += 60*c3p1*c5p2*c7p3*ss[34];
  root_2 += 60*c3p1*c5p3*c7p2*ss[32];
  root_2 += 30*c3p1*c5p4*c7p1*ss[30];
  root_2 += 60*c3p2*c5p1*c7p3*ss[32];
  root_2 += 90*c3p2*c5p2*c7p2*ss[30];
  root_2 += 60*c3p2*c5p3*c7p1*ss[28];
  root_2 += 60*c3p3*c5p1*c7p2*ss[28];
  root_2 += 60*c3p3*c5p2*c7p1*ss[26];
  root_2 += 30*c3p4*c5p1*c7p1*ss[24];
  root_2 += 6*c1p1*c3p5*ss[16];
  root_2 += 6*c1p1*c5p5*ss[26];
  root_2 += 6*c1p1*c7p5*ss[36];
  root_2 += 15*c1p2*c3p4*ss[14];
  root_2 += 15*c1p2*c5p4*ss[22];
  root_2 += 15*c1p2*c7p4*ss[30];
  root_2 += 20*c1p3*c3p3*ss[12];
  root_2 += 20*c1p3*c5p3*ss[18];
  root_2 += 20*c1p3*c7p3*ss[24];
  root_2 += 15*c1p4*c3p2*ss[10];
  root_2 += 15*c1p4*c5p2*ss[14];
  root_2 += 15*c1p4*c7p2*ss[18];
  root_2 += 6*c1p5*c3p1*ss[8];
  root_2 += 6*c1p5*c5p1*ss[10];
  root_2 += 6*c1p5*c7p1*ss[12];
  root_2 += 6*c3p1*c5p5*ss[28];
  root_2 += 6*c3p1*c7p5*ss[38];
  root_2 += 15*c3p2*c5p4*ss[26];
  root_2 += 15*c3p2*c7p4*ss[34];
  root_2 += 20*c3p3*c5p3*ss[24];
  root_2 += 20*c3p3*c7p3*ss[30];
  root_2 += 15*c3p4*c5p2*ss[22];
  root_2 += 15*c3p4*c7p2*ss[26];
  root_2 += 6*c3p5*c5p1*ss[20];
  root_2 += 6*c3p5*c7p1*ss[22];
  root_2 += 6*c5p1*c7p5*ss[40];
  root_2 += 15*c5p2*c7p4*ss[38];
  root_2 += 20*c5p3*c7p3*ss[36];
  root_2 += 15*c5p4*c7p2*ss[34];
  root_2 += 6*c5p5*c7p1*ss[32];
  root_2 += c1p6*ss[6];
  root_2 += c3p6*ss[18];
  root_2 += c5p6*ss[30];
  root_2 += c7p6*ss[42];
  root_2 += -tmom[6];
  T c1p7 = c1p6*c1p1;
  T c1p8 = c1p7*c1p1;
  T c3p7 = c3p6*c3p1;
  T c3p8 = c3p7*c3p1;
  T c5p7 = c5p6*c5p1;
  T c5p8 = c5p7*c5p1;
  T c7p7 = c7p6*c7p1;
  T c7p8 = c7p7*c7p1;
  TACC root_3 = convert<TACC>(336*c1p1*c3p1*c5p1*c7p5*ss[44]);
  root_3 += 840*c1p1*c3p1*c5p2*c7p4*ss[42];
  root_3 += 1120*c1p1*c3p1*c5p3*c7p3*ss[40];
  root_3 += 840*c1p1*c3p1*c5p4*c7p2*ss[38];
  root_3 += 336*c1p1*c3p1*c5p5*c7p1*ss[36];
  root_3 += 840*c1p1*c3p2*c5p1*c7p4*ss[40];
  root_3 += 1680*c1p1*c3p2*c5p2*c7p3*ss[38];
  root_3 += 1680*c1p1*c3p2*c5p3*c7p2*ss[36];
  root_3 += 840*c1p1*c3p2*c5p4*c7p1*ss[34];
  root_3 += 1120*c1p1*c3p3*c5p1*c7p3*ss[36];
  root_3 += 1680*c1p1*c3p3*c5p2*c7p2*ss[34];
  root_3 += 1120*c1p1*c3p3*c5p3*c7p1*ss[32];
  root_3 += 840*c1p1*c3p4*c5p1*c7p2*ss[32];
  root_3 += 840*c1p1*c3p4*c5p2*c7p1*ss[30];
  root_3 += 336*c1p1*c3p5*c5p1*c7p1*ss[28];
  root_3 += 840*c1p2*c3p1*c5p1*c7p4*ss[38];
  root_3 += 1680*c1p2*c3p1*c5p2*c7p3*ss[36];
  root_3 += 1680*c1p2*c3p1*c5p3*c7p2*ss[34];
  root_3 += 840*c1p2*c3p1*c5p4*c7p1*ss[32];
  root_3 += 1680*c1p2*c3p2*c5p1*c7p3*ss[34];
  root_3 += 2520*c1p2*c3p2*c5p2*c7p2*ss[32];
  root_3 += 1680*c1p2*c3p2*c5p3*c7p1*ss[30];
  root_3 += 1680*c1p2*c3p3*c5p1*c7p2*ss[30];
  root_3 += 1680*c1p2*c3p3*c5p2*c7p1*ss[28];
  root_3 += 840*c1p2*c3p4*c5p1*c7p1*ss[26];
  root_3 += 1120*c1p3*c3p1*c5p1*c7p3*ss[32];
  root_3 += 1680*c1p3*c3p1*c5p2*c7p2*ss[30];
  root_3 += 1120*c1p3*c3p1*c5p3*c7p1*ss[28];
  root_3 += 1680*c1p3*c3p2*c5p1*c7p2*ss[28];
  root_3 += 1680*c1p3*c3p2*c5p2*c7p1*ss[26];
  root_3 += 1120*c1p3*c3p3*c5p1*c7p1*ss[24];
  root_3 += 840*c1p4*c3p1*c5p1*c7p2*ss[26];
  root_3 += 840*c1p4*c3p1*c5p2*c7p1*ss[24];
  root_3 += 840*c1p4*c3p2*c5p1*c7p1*ss[22];
  root_3 += 336*c1p5*c3p1*c5p1*c7p1*ss[20];
  root_3 += 56*c1p1*c3p1*c5p6*ss[34];
  root_3 += 56*c1p1*c3p1*c7p6*ss[46];
  root_3 += 168*c1p1*c3p2*c5p5*ss[32];
  root_3 += 168*c1p1*c3p2*c7p5*ss[42];
  root_3 += 280*c1p1*c3p3*c5p4*ss[30];
  root_3 += 280*c1p1*c3p3*c7p4*ss[38];
  root_3 += 280*c1p1*c3p4*c5p3*ss[28];
  root_3 += 280*c1p1*c3p4*c7p3*ss[34];
  root_3 += 168*c1p1*c3p5*c5p2*ss[26];
  root_3 += 168*c1p1*c3p5*c7p2*ss[30];
  root_3 += 56*c1p1*c3p6*c5p1*ss[24];
  root_3 += 56*c1p1*c3p6*c7p1*ss[26];
  root_3 += 56*c1p1*c5p1*c7p6*ss[48];
  root_3 += 168*c1p1*c5p2*c7p5*ss[46];
  root_3 += 280*c1p1*c5p3*c7p4*ss[44];
  root_3 += 280*c1p1*c5p4*c7p3*ss[42];
  root_3 += 168*c1p1*c5p5*c7p2*ss[40];
  root_3 += 56*c1p1*c5p6*c7p1*ss[38];
  root_3 += 168*c1p2*c3p1*c5p5*ss[30];
  root_3 += 168*c1p2*c3p1*c7p5*ss[40];
  root_3 += 420*c1p2*c3p2*c5p4*ss[28];
  root_3 += 420*c1p2*c3p2*c7p4*ss[36];
  root_3 += 560*c1p2*c3p3*c5p3*ss[26];
  root_3 += 560*c1p2*c3p3*c7p3*ss[32];
  root_3 += 420*c1p2*c3p4*c5p2*ss[24];
  root_3 += 420*c1p2*c3p4*c7p2*ss[28];
  root_3 += 168*c1p2*c3p5*c5p1*ss[22];
  root_3 += 168*c1p2*c3p5*c7p1*ss[24];
  root_3 += 168*c1p2*c5p1*c7p5*ss[42];
  root_3 += 420*c1p2*c5p2*c7p4*ss[40];
  root_3 += 560*c1p2*c5p3*c7p3*ss[38];
  root_3 += 420*c1p2*c5p4*c7p2*ss[36];
  root_3 += 168*c1p2*c5p5*c7p1*ss[34];
  root_3 += 280*c1p3*c3p1*c5p4*ss[26];
  root_3 += 280*c1p3*c3p1*c7p4*ss[34];
  root_3 += 560*c1p3*c3p2*c5p3*ss[24];
  root_3 += 560*c1p3*c3p2*c7p3*ss[30];
  root_3 += 560*c1p3*c3p3*c5p2*ss[22];
  root_3 += 560*c1p3*c3p3*c7p2*ss[26];
  root_3 += 280*c1p3*c3p4*c5p1*ss[20];
  root_3 += 280*c1p3*c3p4*c7p1*ss[22];
  root_3 += 280*c1p3*c5p1*c7p4*ss[36];
  root_3 += 560*c1p3*c5p2*c7p3*ss[34];
  root_3 += 560*c1p3*c5p3*c7p2*ss[32];
  root_3 += 280*c1p3*c5p4*c7p1*ss[30];
  root_3 += 280*c1p4*c3p1*c5p3*ss[22];
  root_3 += 280*c1p4*c3p1*c7p3*ss[28];
  root_3 += 420*c1p4*c3p2*c5p2*ss[20];
  root_3 += 420*c1p4*c3p2*c7p2*ss[24];
  root_3 += 280*c1p4*c3p3*c5p1*ss[18];
  root_3 += 280*c1p4*c3p3*c7p1*ss[20];
  root_3 += 280*c1p4*c5p1*c7p3*ss[30];
  root_3 += 420*c1p4*c5p2*c7p2*ss[28];
  root_3 += 280*c1p4*c5p3*c7p1*ss[26];
  root_3 += 168*c1p5*c3p1*c5p2*ss[18];
  root_3 += 168*c1p5*c3p1*c7p2*ss[22];
  root_3 += 168*c1p5*c3p2*c5p1*ss[16];
  root_3 += 168*c1p5*c3p2*c7p1*ss[18];
  root_3 += 168*c1p5*c5p1*c7p2*ss[24];
  root_3 += 168*c1p5*c5p2*c7p1*ss[22];
  root_3 += 56*c1p6*c3p1*c5p1*ss[14];
  root_3 += 56*c1p6*c3p1*c7p1*ss[16];
  root_3 += 56*c1p6*c5p1*c7p1*ss[18];
  root_3 += 56*c3p1*c5p1*c7p6*ss[50];
  root_3 += 168*c3p1*c5p2*c7p5*ss[48];
  root_3 += 280*c3p1*c5p3*c7p4*ss[46];
  root_3 += 280*c3p1*c5p4*c7p3*ss[44];
  root_3 += 168*c3p1*c5p5*c7p2*ss[42];
  root_3 += 56*c3p1*c5p6*c7p1*ss[40];
  root_3 += 168*c3p2*c5p1*c7p5*ss[46];
  root_3 += 420*c3p2*c5p2*c7p4*ss[44];
  root_3 += 560*c3p2*c5p3*c7p3*ss[42];
  root_3 += 420*c3p2*c5p4*c7p2*ss[40];
  root_3 += 168*c3p2*c5p5*c7p1*ss[38];
  root_3 += 280*c3p3*c5p1*c7p4*ss[42];
  root_3 += 560*c3p3*c5p2*c7p3*ss[40];
  root_3 += 560*c3p3*c5p3*c7p2*ss[38];
  root_3 += 280*c3p3*c5p4*c7p1*ss[36];
  root_3 += 280*c3p4*c5p1*c7p3*ss[38];
  root_3 += 420*c3p4*c5p2*c7p2*ss[36];
  root_3 += 280*c3p4*c5p3*c7p1*ss[34];
  root_3 += 168*c3p5*c5p1*c7p2*ss[34];
  root_3 += 168*c3p5*c5p2*c7p1*ss[32];
  root_3 += 56*c3p6*c5p1*c7p1*ss[30];
  root_3 += 8*c1p1*c3p7*ss[22];
  root_3 += 8*c1p1*c5p7*ss[36];
  root_3 += 8*c1p1*c7p7*ss[50];
  root_3 += 28*c1p2*c3p6*ss[20];
  root_3 += 28*c1p2*c5p6*ss[32];
  root_3 += 28*c1p2*c7p6*ss[44];
  root_3 += 56*c1p3*c3p5*ss[18];
  root_3 += 56*c1p3*c5p5*ss[28];
  root_3 += 56*c1p3*c7p5*ss[38];
  root_3 += 70*c1p4*c3p4*ss[16];
  root_3 += 70*c1p4*c5p4*ss[24];
  root_3 += 70*c1p4*c7p4*ss[32];
  root_3 += 56*c1p5*c3p3*ss[14];
  root_3 += 56*c1p5*c5p3*ss[20];
  root_3 += 56*c1p5*c7p3*ss[26];
  root_3 += 28*c1p6*c3p2*ss[12];
  root_3 += 28*c1p6*c5p2*ss[16];
  root_3 += 28*c1p6*c7p2*ss[20];
  root_3 += 8*c1p7*c3p1*ss[10];
  root_3 += 8*c1p7*c5p1*ss[12];
  root_3 += 8*c1p7*c7p1*ss[14];
  root_3 += 8*c3p1*c5p7*ss[38];
  root_3 += 8*c3p1*c7p7*ss[52];
  root_3 += 28*c3p2*c5p6*ss[36];
  root_3 += 28*c3p2*c7p6*ss[48];
  root_3 += 56*c3p3*c5p5*ss[34];
  root_3 += 56*c3p3*c7p5*ss[44];
  root_3 += 70*c3p4*c5p4*ss[32];
  root_3 += 70*c3p4*c7p4*ss[40];
  root_3 += 56*c3p5*c5p3*ss[30];
  root_3 += 56*c3p5*c7p3*ss[36];
  root_3 += 28*c3p6*c5p2*ss[28];
  root_3 += 28*c3p6*c7p2*ss[32];
  root_3 += 8*c3p7*c5p1*ss[26];
  root_3 += 8*c3p7*c7p1*ss[28];
  root_3 += 8*c5p1*c7p7*ss[54];
  root_3 += 28*c5p2*c7p6*ss[52];
  root_3 += 56*c5p3*c7p5*ss[50];
  root_3 += 70*c5p4*c7p4*ss[48];
  root_3 += 56*c5p5*c7p3*ss[46];
  root_3 += 28*c5p6*c7p2*ss[44];
  root_3 += 8*c5p7*c7p1*ss[42];
  root_3 += c1p8*ss[8];
  root_3 += c3p8*ss[24];
  root_3 += c5p8*ss[40];
  root_3 += c7p8*ss[56];
  root_3 += -tmom[8];
  TACC jac_0_0 = convert<TACC>(2*c1p1*ss[2]);
  jac_0_0 += 2*c3p1*ss[4];
  jac_0_0 += 2*c5p1*ss[6];
  jac_0_0 += 2*c7p1*ss[8];
  TACC jac_0_1 = convert<TACC>(2*c1p1*ss[4]);
  jac_0_1 += 2*c3p1*ss[6];
  jac_0_1 += 2*c5p1*ss[8];
  jac_0_1 += 2*c7p1*ss[10];
  TACC jac_0_2 = convert<TACC>(2*c1p1*ss[6]);
  jac_0_2 += 2*c3p1*ss[8];
  jac_0_2 += 2*c5p1*ss[10];
  jac_0_2 += 2*c7p1*ss[12];
  TACC jac_0_3 = convert<TACC>(2*c1p1*ss[8]);
  jac_0_3 += 2*c3p1*ss[10];
  jac_0_3 += 2*c5p1*ss[12];
  jac_0_3 += 2*c7p1*ss[14];
  TACC jac_1_0 = convert<TACC>(24*c1p1*c3p1*c5p1*ss[10]);
  jac_1_0 += 24*c1p1*c3p1*c7p1*ss[12];
  jac_1_0 += 24*c1p1*c5p1*c7p1*ss[14];
  jac_1_0 += 24*c3p1*c5p1*c7p1*ss[16];
  jac_1_0 += 12*c1p1*c3p2*ss[8];
  jac_1_0 += 12*c1p1*c5p2*ss[12];
  jac_1_0 += 12*c1p1*c7p2*ss[16];
  jac_1_0 += 12*c1p2*c3p1*ss[6];
  jac_1_0 += 12*c1p2*c5p1*ss[8];
  jac_1_0 += 12*c1p2*c7p1*ss[10];
  jac_1_0 += 12*c3p1*c5p2*ss[14];
  jac_1_0 += 12*c3p1*c7p2*ss[18];
  jac_1_0 += 12*c3p2*c5p1*ss[12];
  jac_1_0 += 12*c3p2*c7p1*ss[14];
  jac_1_0 += 12*c5p1*c7p2*ss[20];
  jac_1_0 += 12*c5p2*c7p1*ss[18];
  jac_1_0 += 4*c1p3*ss[4];
  jac_1_0 += 4*c3p3*ss[10];
  jac_1_0 += 4*c5p3*ss[16];
  jac_1_0 += 4*c7p3*ss[22];
  TACC jac_1_1 = convert<TACC>(24*c1p1*c3p1*c5p1*ss[12]);
  jac_1_1 += 24*c1p1*c3p1*c7p1*ss[14];
  jac_1_1 += 24*c1p1*c5p1*c7p1*ss[16];
  jac_1_1 += 24*c3p1*c5p1*c7p1*ss[18];
  jac_1_1 += 12*c1p1*c3p2*ss[10];
  jac_1_1 += 12*c1p1*c5p2*ss[14];
  jac_1_1 += 12*c1p1*c7p2*ss[18];
  jac_1_1 += 12*c1p2*c3p1*ss[8];
  jac_1_1 += 12*c1p2*c5p1*ss[10];
  jac_1_1 += 12*c1p2*c7p1*ss[12];
  jac_1_1 += 12*c3p1*c5p2*ss[16];
  jac_1_1 += 12*c3p1*c7p2*ss[20];
  jac_1_1 += 12*c3p2*c5p1*ss[14];
  jac_1_1 += 12*c3p2*c7p1*ss[16];
  jac_1_1 += 12*c5p1*c7p2*ss[22];
  jac_1_1 += 12*c5p2*c7p1*ss[20];
  jac_1_1 += 4*c1p3*ss[6];
  jac_1_1 += 4*c3p3*ss[12];
  jac_1_1 += 4*c5p3*ss[18];
  jac_1_1 += 4*c7p3*ss[24];
  TACC jac_1_2 = convert<TACC>(24*c1p1*c3p1*c5p1*ss[14]);
  jac_1_2 += 24*c1p1*c3p1*c7p1*ss[16];
  jac_1_2 += 24*c1p1*c5p1*c7p1*ss[18];
  jac_1_2 += 24*c3p1*c5p1*c7p1*ss[20];
  jac_1_2 += 12*c1p1*c3p2*ss[12];
  jac_1_2 += 12*c1p1*c5p2*ss[16];
  jac_1_2 += 12*c1p1*c7p2*ss[20];
  jac_1_2 += 12*c1p2*c3p1*ss[10];
  jac_1_2 += 12*c1p2*c5p1*ss[12];
  jac_1_2 += 12*c1p2*c7p1*ss[14];
  jac_1_2 += 12*c3p1*c5p2*ss[18];
  jac_1_2 += 12*c3p1*c7p2*ss[22];
  jac_1_2 += 12*c3p2*c5p1*ss[16];
  jac_1_2 += 12*c3p2*c7p1*ss[18];
  jac_1_2 += 12*c5p1*c7p2*ss[24];
  jac_1_2 += 12*c5p2*c7p1*ss[22];
  jac_1_2 += 4*c1p3*ss[8];
  jac_1_2 += 4*c3p3*ss[14];
  jac_1_2 += 4*c5p3*ss[20];
  jac_1_2 += 4*c7p3*ss[26];
  TACC jac_1_3 = convert<TACC>(24*c1p1*c3p1*c5p1*ss[16]);
  jac_1_3 += 24*c1p1*c3p1*c7p1*ss[18];
  jac_1_3 += 24*c1p1*c5p1*c7p1*ss[20];
  jac_1_3 += 24*c3p1*c5p1*c7p1*ss[22];
  jac_1_3 += 12*c1p1*c3p2*ss[14];
  jac_1_3 += 12*c1p1*c5p2*ss[18];
  jac_1_3 += 12*c1p1*c7p2*ss[22];
  jac_1_3 += 12*c1p2*c3p1*ss[12];
  jac_1_3 += 12*c1p2*c5p1*ss[14];
  jac_1_3 += 12*c1p2*c7p1*ss[16];
  jac_1_3 += 12*c3p1*c5p2*ss[20];
  jac_1_3 += 12*c3p1*c7p2*ss[24];
  jac_1_3 += 12*c3p2*c5p1*ss[18];
  jac_1_3 += 12*c3p2*c7p1*ss[20];
  jac_1_3 += 12*c5p1*c7p2*ss[26];
  jac_1_3 += 12*c5p2*c7p1*ss[24];
  jac_1_3 += 4*c1p3*ss[10];
  jac_1_3 += 4*c3p3*ss[16];
  jac_1_3 += 4*c5p3*ss[22];
  jac_1_3 += 4*c7p3*ss[28];
  TACC jac_2_0 = convert<TACC>(360*c1p1*c3p1*c5p1*c7p2*ss[24]);
  jac_2_0 += 360*c1p1*c3p1*c5p2*c7p1*ss[22];
  jac_2_0 += 360*c1p1*c3p2*c5p1*c7p1*ss[20];
  jac_2_0 += 360*c1p2*c3p1*c5p1*c7p1*ss[18];
  jac_2_0 += 120*c1p1*c3p1*c5p3*ss[20];
  jac_2_0 += 120*c1p1*c3p1*c7p3*ss[26];
  jac_2_0 += 180*c1p1*c3p2*c5p2*ss[18];
  jac_2_0 += 180*c1p1*c3p2*c7p2*ss[22];
  jac_2_0 += 120*c1p1*c3p3*c5p1*ss[16];
  jac_2_0 += 120*c1p1*c3p3*c7p1*ss[18];
  jac_2_0 += 120*c1p1*c5p1*c7p3*ss[28];
  jac_2_0 += 180*c1p1*c5p2*c7p2*ss[26];
  jac_2_0 += 120*c1p1*c5p3*c7p1*ss[24];
  jac_2_0 += 180*c1p2*c3p1*c5p2*ss[16];
  jac_2_0 += 180*c1p2*c3p1*c7p2*ss[20];
  jac_2_0 += 180*c1p2*c3p2*c5p1*ss[14];
  jac_2_0 += 180*c1p2*c3p2*c7p1*ss[16];
  jac_2_0 += 180*c1p2*c5p1*c7p2*ss[22];
  jac_2_0 += 180*c1p2*c5p2*c7p1*ss[20];
  jac_2_0 += 120*c1p3*c3p1*c5p1*ss[12];
  jac_2_0 += 120*c1p3*c3p1*c7p1*ss[14];
  jac_2_0 += 120*c1p3*c5p1*c7p1*ss[16];
  jac_2_0 += 120*c3p1*c5p1*c7p3*ss[30];
  jac_2_0 += 180*c3p1*c5p2*c7p2*ss[28];
  jac_2_0 += 120*c3p1*c5p3*c7p1*ss[26];
  jac_2_0 += 180*c3p2*c5p1*c7p2*ss[26];
  jac_2_0 += 180*c3p2*c5p2*c7p1*ss[24];
  jac_2_0 += 120*c3p3*c5p1*c7p1*ss[22];
  jac_2_0 += 30*c1p1*c3p4*ss[14];
  jac_2_0 += 30*c1p1*c5p4*ss[22];
  jac_2_0 += 30*c1p1*c7p4*ss[30];
  jac_2_0 += 60*c1p2*c3p3*ss[12];
  jac_2_0 += 60*c1p2*c5p3*ss[18];
  jac_2_0 += 60*c1p2*c7p3*ss[24];
  jac_2_0 += 60*c1p3*c3p2*ss[10];
  jac_2_0 += 60*c1p3*c5p2*ss[14];
  jac_2_0 += 60*c1p3*c7p2*ss[18];
  jac_2_0 += 30*c1p4*c3p1*ss[8];
  jac_2_0 += 30*c1p4*c5p1*ss[10];
  jac_2_0 += 30*c1p4*c7p1*ss[12];
  jac_2_0 += 30*c3p1*c5p4*ss[24];
  jac_2_0 += 30*c3p1*c7p4*ss[32];
  jac_2_0 += 60*c3p2*c5p3*ss[22];
  jac_2_0 += 60*c3p2*c7p3*ss[28];
  jac_2_0 += 60*c3p3*c5p2*ss[20];
  jac_2_0 += 60*c3p3*c7p2*ss[24];
  jac_2_0 += 30*c3p4*c5p1*ss[18];
  jac_2_0 += 30*c3p4*c7p1*ss[20];
  jac_2_0 += 30*c5p1*c7p4*ss[34];
  jac_2_0 += 60*c5p2*c7p3*ss[32];
  jac_2_0 += 60*c5p3*c7p2*ss[30];
  jac_2_0 += 30*c5p4*c7p1*ss[28];
  jac_2_0 += 6*c1p5*ss[6];
  jac_2_0 += 6*c3p5*ss[16];
  jac_2_0 += 6*c5p5*ss[26];
  jac_2_0 += 6*c7p5*ss[36];
  TACC jac_2_1 = convert<TACC>(360*c1p1*c3p1*c5p1*c7p2*ss[26]);
  jac_2_1 += 360*c1p1*c3p1*c5p2*c7p1*ss[24];
  jac_2_1 += 360*c1p1*c3p2*c5p1*c7p1*ss[22];
  jac_2_1 += 360*c1p2*c3p1*c5p1*c7p1*ss[20];
  jac_2_1 += 120*c1p1*c3p1*c5p3*ss[22];
  jac_2_1 += 120*c1p1*c3p1*c7p3*ss[28];
  jac_2_1 += 180*c1p1*c3p2*c5p2*ss[20];
  jac_2_1 += 180*c1p1*c3p2*c7p2*ss[24];
  jac_2_1 += 120*c1p1*c3p3*c5p1*ss[18];
  jac_2_1 += 120*c1p1*c3p3*c7p1*ss[20];
  jac_2_1 += 120*c1p1*c5p1*c7p3*ss[30];
  jac_2_1 += 180*c1p1*c5p2*c7p2*ss[28];
  jac_2_1 += 120*c1p1*c5p3*c7p1*ss[26];
  jac_2_1 += 180*c1p2*c3p1*c5p2*ss[18];
  jac_2_1 += 180*c1p2*c3p1*c7p2*ss[22];
  jac_2_1 += 180*c1p2*c3p2*c5p1*ss[16];
  jac_2_1 += 180*c1p2*c3p2*c7p1*ss[18];
  jac_2_1 += 180*c1p2*c5p1*c7p2*ss[24];
  jac_2_1 += 180*c1p2*c5p2*c7p1*ss[22];
  jac_2_1 += 120*c1p3*c3p1*c5p1*ss[14];
  jac_2_1 += 120*c1p3*c3p1*c7p1*ss[16];
  jac_2_1 += 120*c1p3*c5p1*c7p1*ss[18];
  jac_2_1 += 120*c3p1*c5p1*c7p3*ss[32];
  jac_2_1 += 180*c3p1*c5p2*c7p2*ss[30];
  jac_2_1 += 120*c3p1*c5p3*c7p1*ss[28];
  jac_2_1 += 180*c3p2*c5p1*c7p2*ss[28];
  jac_2_1 += 180*c3p2*c5p2*c7p1*ss[26];
  jac_2_1 += 120*c3p3*c5p1*c7p1*ss[24];
  jac_2_1 += 30*c1p1*c3p4*ss[16];
  jac_2_1 += 30*c1p1*c5p4*ss[24];
  jac_2_1 += 30*c1p1*c7p4*ss[32];
  jac_2_1 += 60*c1p2*c3p3*ss[14];
  jac_2_1 += 60*c1p2*c5p3*ss[20];
  jac_2_1 += 60*c1p2*c7p3*ss[26];
  jac_2_1 += 60*c1p3*c3p2*ss[12];
  jac_2_1 += 60*c1p3*c5p2*ss[16];
  jac_2_1 += 60*c1p3*c7p2*ss[20];
  jac_2_1 += 30*c1p4*c3p1*ss[10];
  jac_2_1 += 30*c1p4*c5p1*ss[12];
  jac_2_1 += 30*c1p4*c7p1*ss[14];
  jac_2_1 += 30*c3p1*c5p4*ss[26];
  jac_2_1 += 30*c3p1*c7p4*ss[34];
  jac_2_1 += 60*c3p2*c5p3*ss[24];
  jac_2_1 += 60*c3p2*c7p3*ss[30];
  jac_2_1 += 60*c3p3*c5p2*ss[22];
  jac_2_1 += 60*c3p3*c7p2*ss[26];
  jac_2_1 += 30*c3p4*c5p1*ss[20];
  jac_2_1 += 30*c3p4*c7p1*ss[22];
  jac_2_1 += 30*c5p1*c7p4*ss[36];
  jac_2_1 += 60*c5p2*c7p3*ss[34];
  jac_2_1 += 60*c5p3*c7p2*ss[32];
  jac_2_1 += 30*c5p4*c7p1*ss[30];
  jac_2_1 += 6*c1p5*ss[8];
  jac_2_1 += 6*c3p5*ss[18];
  jac_2_1 += 6*c5p5*ss[28];
  jac_2_1 += 6*c7p5*ss[38];
  TACC jac_2_2 = convert<TACC>(360*c1p1*c3p1*c5p1*c7p2*ss[28]);
  jac_2_2 += 360*c1p1*c3p1*c5p2*c7p1*ss[26];
  jac_2_2 += 360*c1p1*c3p2*c5p1*c7p1*ss[24];
  jac_2_2 += 360*c1p2*c3p1*c5p1*c7p1*ss[22];
  jac_2_2 += 120*c1p1*c3p1*c5p3*ss[24];
  jac_2_2 += 120*c1p1*c3p1*c7p3*ss[30];
  jac_2_2 += 180*c1p1*c3p2*c5p2*ss[22];
  jac_2_2 += 180*c1p1*c3p2*c7p2*ss[26];
  jac_2_2 += 120*c1p1*c3p3*c5p1*ss[20];
  jac_2_2 += 120*c1p1*c3p3*c7p1*ss[22];
  jac_2_2 += 120*c1p1*c5p1*c7p3*ss[32];
  jac_2_2 += 180*c1p1*c5p2*c7p2*ss[30];
  jac_2_2 += 120*c1p1*c5p3*c7p1*ss[28];
  jac_2_2 += 180*c1p2*c3p1*c5p2*ss[20];
  jac_2_2 += 180*c1p2*c3p1*c7p2*ss[24];
  jac_2_2 += 180*c1p2*c3p2*c5p1*ss[18];
  jac_2_2 += 180*c1p2*c3p2*c7p1*ss[20];
  jac_2_2 += 180*c1p2*c5p1*c7p2*ss[26];
  jac_2_2 += 180*c1p2*c5p2*c7p1*ss[24];
  jac_2_2 += 120*c1p3*c3p1*c5p1*ss[16];
  jac_2_2 += 120*c1p3*c3p1*c7p1*ss[18];
  jac_2_2 += 120*c1p3*c5p1*c7p1*ss[20];
  jac_2_2 += 120*c3p1*c5p1*c7p3*ss[34];
  jac_2_2 += 180*c3p1*c5p2*c7p2*ss[32];
  jac_2_2 += 120*c3p1*c5p3*c7p1*ss[30];
  jac_2_2 += 180*c3p2*c5p1*c7p2*ss[30];
  jac_2_2 += 180*c3p2*c5p2*c7p1*ss[28];
  jac_2_2 += 120*c3p3*c5p1*c7p1*ss[26];
  jac_2_2 += 30*c1p1*c3p4*ss[18];
  jac_2_2 += 30*c1p1*c5p4*ss[26];
  jac_2_2 += 30*c1p1*c7p4*ss[34];
  jac_2_2 += 60*c1p2*c3p3*ss[16];
  jac_2_2 += 60*c1p2*c5p3*ss[22];
  jac_2_2 += 60*c1p2*c7p3*ss[28];
  jac_2_2 += 60*c1p3*c3p2*ss[14];
  jac_2_2 += 60*c1p3*c5p2*ss[18];
  jac_2_2 += 60*c1p3*c7p2*ss[22];
  jac_2_2 += 30*c1p4*c3p1*ss[12];
  jac_2_2 += 30*c1p4*c5p1*ss[14];
  jac_2_2 += 30*c1p4*c7p1*ss[16];
  jac_2_2 += 30*c3p1*c5p4*ss[28];
  jac_2_2 += 30*c3p1*c7p4*ss[36];
  jac_2_2 += 60*c3p2*c5p3*ss[26];
  jac_2_2 += 60*c3p2*c7p3*ss[32];
  jac_2_2 += 60*c3p3*c5p2*ss[24];
  jac_2_2 += 60*c3p3*c7p2*ss[28];
  jac_2_2 += 30*c3p4*c5p1*ss[22];
  jac_2_2 += 30*c3p4*c7p1*ss[24];
  jac_2_2 += 30*c5p1*c7p4*ss[38];
  jac_2_2 += 60*c5p2*c7p3*ss[36];
  jac_2_2 += 60*c5p3*c7p2*ss[34];
  jac_2_2 += 30*c5p4*c7p1*ss[32];
  jac_2_2 += 6*c1p5*ss[10];
  jac_2_2 += 6*c3p5*ss[20];
  jac_2_2 += 6*c5p5*ss[30];
  jac_2_2 += 6*c7p5*ss[40];
  TACC jac_2_3 = convert<TACC>(360*c1p1*c3p1*c5p1*c7p2*ss[30]);
  jac_2_3 += 360*c1p1*c3p1*c5p2*c7p1*ss[28];
  jac_2_3 += 360*c1p1*c3p2*c5p1*c7p1*ss[26];
  jac_2_3 += 360*c1p2*c3p1*c5p1*c7p1*ss[24];
  jac_2_3 += 120*c1p1*c3p1*c5p3*ss[26];
  jac_2_3 += 120*c1p1*c3p1*c7p3*ss[32];
  jac_2_3 += 180*c1p1*c3p2*c5p2*ss[24];
  jac_2_3 += 180*c1p1*c3p2*c7p2*ss[28];
  jac_2_3 += 120*c1p1*c3p3*c5p1*ss[22];
  jac_2_3 += 120*c1p1*c3p3*c7p1*ss[24];
  jac_2_3 += 120*c1p1*c5p1*c7p3*ss[34];
  jac_2_3 += 180*c1p1*c5p2*c7p2*ss[32];
  jac_2_3 += 120*c1p1*c5p3*c7p1*ss[30];
  jac_2_3 += 180*c1p2*c3p1*c5p2*ss[22];
  jac_2_3 += 180*c1p2*c3p1*c7p2*ss[26];
  jac_2_3 += 180*c1p2*c3p2*c5p1*ss[20];
  jac_2_3 += 180*c1p2*c3p2*c7p1*ss[22];
  jac_2_3 += 180*c1p2*c5p1*c7p2*ss[28];
  jac_2_3 += 180*c1p2*c5p2*c7p1*ss[26];
  jac_2_3 += 120*c1p3*c3p1*c5p1*ss[18];
  jac_2_3 += 120*c1p3*c3p1*c7p1*ss[20];
  jac_2_3 += 120*c1p3*c5p1*c7p1*ss[22];
  jac_2_3 += 120*c3p1*c5p1*c7p3*ss[36];
  jac_2_3 += 180*c3p1*c5p2*c7p2*ss[34];
  jac_2_3 += 120*c3p1*c5p3*c7p1*ss[32];
  jac_2_3 += 180*c3p2*c5p1*c7p2*ss[32];
  jac_2_3 += 180*c3p2*c5p2*c7p1*ss[30];
  jac_2_3 += 120*c3p3*c5p1*c7p1*ss[28];
  jac_2_3 += 30*c1p1*c3p4*ss[20];
  jac_2_3 += 30*c1p1*c5p4*ss[28];
  jac_2_3 += 30*c1p1*c7p4*ss[36];
  jac_2_3 += 60*c1p2*c3p3*ss[18];
  jac_2_3 += 60*c1p2*c5p3*ss[24];
  jac_2_3 += 60*c1p2*c7p3*ss[30];
  jac_2_3 += 60*c1p3*c3p2*ss[16];
  jac_2_3 += 60*c1p3*c5p2*ss[20];
  jac_2_3 += 60*c1p3*c7p2*ss[24];
  jac_2_3 += 30*c1p4*c3p1*ss[14];
  jac_2_3 += 30*c1p4*c5p1*ss[16];
  jac_2_3 += 30*c1p4*c7p1*ss[18];
  jac_2_3 += 30*c3p1*c5p4*ss[30];
  jac_2_3 += 30*c3p1*c7p4*ss[38];
  jac_2_3 += 60*c3p2*c5p3*ss[28];
  jac_2_3 += 60*c3p2*c7p3*ss[34];
  jac_2_3 += 60*c3p3*c5p2*ss[26];
  jac_2_3 += 60*c3p3*c7p2*ss[30];
  jac_2_3 += 30*c3p4*c5p1*ss[24];
  jac_2_3 += 30*c3p4*c7p1*ss[26];
  jac_2_3 += 30*c5p1*c7p4*ss[40];
  jac_2_3 += 60*c5p2*c7p3*ss[38];
  jac_2_3 += 60*c5p3*c7p2*ss[36];
  jac_2_3 += 30*c5p4*c7p1*ss[34];
  jac_2_3 += 6*c1p5*ss[12];
  jac_2_3 += 6*c3p5*ss[22];
  jac_2_3 += 6*c5p5*ss[32];
  jac_2_3 += 6*c7p5*ss[42];
  TACC jac_3_0 = convert<TACC>(1680*c1p1*c3p1*c5p1*c7p4*ss[38]);
  jac_3_0 += 3360*c1p1*c3p1*c5p2*c7p3*ss[36];
  jac_3_0 += 3360*c1p1*c3p1*c5p3*c7p2*ss[34];
  jac_3_0 += 1680*c1p1*c3p1*c5p4*c7p1*ss[32];
  jac_3_0 += 3360*c1p1*c3p2*c5p1*c7p3*ss[34];
  jac_3_0 += 5040*c1p1*c3p2*c5p2*c7p2*ss[32];
  jac_3_0 += 3360*c1p1*c3p2*c5p3*c7p1*ss[30];
  jac_3_0 += 3360*c1p1*c3p3*c5p1*c7p2*ss[30];
  jac_3_0 += 3360*c1p1*c3p3*c5p2*c7p1*ss[28];
  jac_3_0 += 1680*c1p1*c3p4*c5p1*c7p1*ss[26];
  jac_3_0 += 3360*c1p2*c3p1*c5p1*c7p3*ss[32];
  jac_3_0 += 5040*c1p2*c3p1*c5p2*c7p2*ss[30];
  jac_3_0 += 3360*c1p2*c3p1*c5p3*c7p1*ss[28];
  jac_3_0 += 5040*c1p2*c3p2*c5p1*c7p2*ss[28];
  jac_3_0 += 5040*c1p2*c3p2*c5p2*c7p1*ss[26];
  jac_3_0 += 3360*c1p2*c3p3*c5p1*c7p1*ss[24];
  jac_3_0 += 3360*c1p3*c3p1*c5p1*c7p2*ss[26];
  jac_3_0 += 3360*c1p3*c3p1*c5p2*c7p1*ss[24];
  jac_3_0 += 3360*c1p3*c3p2*c5p1*c7p1*ss[22];
  jac_3_0 += 1680*c1p4*c3p1*c5p1*c7p1*ss[20];
  jac_3_0 += 336*c1p1*c3p1*c5p5*ss[30];
  jac_3_0 += 336*c1p1*c3p1*c7p5*ss[40];
  jac_3_0 += 840*c1p1*c3p2*c5p4*ss[28];
  jac_3_0 += 840*c1p1*c3p2*c7p4*ss[36];
  jac_3_0 += 1120*c1p1*c3p3*c5p3*ss[26];
  jac_3_0 += 1120*c1p1*c3p3*c7p3*ss[32];
  jac_3_0 += 840*c1p1*c3p4*c5p2*ss[24];
  jac_3_0 += 840*c1p1*c3p4*c7p2*ss[28];
  jac_3_0 += 336*c1p1*c3p5*c5p1*ss[22];
  jac_3_0 += 336*c1p1*c3p5*c7p1*ss[24];
  jac_3_0 += 336*c1p1*c5p1*c7p5*ss[42];
  jac_3_0 += 840*c1p1*c5p2*c7p4*ss[40];
  jac_3_0 += 1120*c1p1*c5p3*c7p3*ss[38];
  jac_3_0 += 840*c1p1*c5p4*c7p2*ss[36];
  jac_3_0 += 336*c1p1*c5p5*c7p1*ss[34];
  jac_3_0 += 840*c1p2*c3p1*c5p4*ss[26];
  jac_3_0 += 840*c1p2*c3p1*c7p4*ss[34];
  jac_3_0 += 1680*c1p2*c3p2*c5p3*ss[24];
  jac_3_0 += 1680*c1p2*c3p2*c7p3*ss[30];
  jac_3_0 += 1680*c1p2*c3p3*c5p2*ss[22];
  jac_3_0 += 1680*c1p2*c3p3*c7p2*ss[26];
  jac_3_0 += 840*c1p2*c3p4*c5p1*ss[20];
  jac_3_0 += 840*c1p2*c3p4*c7p1*ss[22];
  jac_3_0 += 840*c1p2*c5p1*c7p4*ss[36];
  jac_3_0 += 1680*c1p2*c5p2*c7p3*ss[34];
  jac_3_0 += 1680*c1p2*c5p3*c7p2*ss[32];
  jac_3_0 += 840*c1p2*c5p4*c7p1*ss[30];
  jac_3_0 += 1120*c1p3*c3p1*c5p3*ss[22];
  jac_3_0 += 1120*c1p3*c3p1*c7p3*ss[28];
  jac_3_0 += 1680*c1p3*c3p2*c5p2*ss[20];
  jac_3_0 += 1680*c1p3*c3p2*c7p2*ss[24];
  jac_3_0 += 1120*c1p3*c3p3*c5p1*ss[18];
  jac_3_0 += 1120*c1p3*c3p3*c7p1*ss[20];
  jac_3_0 += 1120*c1p3*c5p1*c7p3*ss[30];
  jac_3_0 += 1680*c1p3*c5p2*c7p2*ss[28];
  jac_3_0 += 1120*c1p3*c5p3*c7p1*ss[26];
  jac_3_0 += 840*c1p4*c3p1*c5p2*ss[18];
  jac_3_0 += 840*c1p4*c3p1*c7p2*ss[22];
  jac_3_0 += 840*c1p4*c3p2*c5p1*ss[16];
  jac_3_0 += 840*c1p4*c3p2*c7p1*ss[18];
  jac_3_0 += 840*c1p4*c5p1*c7p2*ss[24];
  jac_3_0 += 840*c1p4*c5p2*c7p1*ss[22];
  jac_3_0 += 336*c1p5*c3p1*c5p1*ss[14];
  jac_3_0 += 336*c1p5*c3p1*c7p1*ss[16];
  jac_3_0 += 336*c1p5*c5p1*c7p1*ss[18];
  jac_3_0 += 336*c3p1*c5p1*c7p5*ss[44];
  jac_3_0 += 840*c3p1*c5p2*c7p4*ss[42];
  jac_3_0 += 1120*c3p1*c5p3*c7p3*ss[40];
  jac_3_0 += 840*c3p1*c5p4*c7p2*ss[38];
  jac_3_0 += 336*c3p1*c5p5*c7p1*ss[36];
  jac_3_0 += 840*c3p2*c5p1*c7p4*ss[40];
  jac_3_0 += 1680*c3p2*c5p2*c7p3*ss[38];
  jac_3_0 += 1680*c3p2*c5p3*c7p2*ss[36];
  jac_3_0 += 840*c3p2*c5p4*c7p1*ss[34];
  jac_3_0 += 1120*c3p3*c5p1*c7p3*ss[36];
  jac_3_0 += 1680*c3p3*c5p2*c7p2*ss[34];
  jac_3_0 += 1120*c3p3*c5p3*c7p1*ss[32];
  jac_3_0 += 840*c3p4*c5p1*c7p2*ss[32];
  jac_3_0 += 840*c3p4*c5p2*c7p1*ss[30];
  jac_3_0 += 336*c3p5*c5p1*c7p1*ss[28];
  jac_3_0 += 56*c1p1*c3p6*ss[20];
  jac_3_0 += 56*c1p1*c5p6*ss[32];
  jac_3_0 += 56*c1p1*c7p6*ss[44];
  jac_3_0 += 168*c1p2*c3p5*ss[18];
  jac_3_0 += 168*c1p2*c5p5*ss[28];
  jac_3_0 += 168*c1p2*c7p5*ss[38];
  jac_3_0 += 280*c1p3*c3p4*ss[16];
  jac_3_0 += 280*c1p3*c5p4*ss[24];
  jac_3_0 += 280*c1p3*c7p4*ss[32];
  jac_3_0 += 280*c1p4*c3p3*ss[14];
  jac_3_0 += 280*c1p4*c5p3*ss[20];
  jac_3_0 += 280*c1p4*c7p3*ss[26];
  jac_3_0 += 168*c1p5*c3p2*ss[12];
  jac_3_0 += 168*c1p5*c5p2*ss[16];
  jac_3_0 += 168*c1p5*c7p2*ss[20];
  jac_3_0 += 56*c1p6*c3p1*ss[10];
  jac_3_0 += 56*c1p6*c5p1*ss[12];
  jac_3_0 += 56*c1p6*c7p1*ss[14];
  jac_3_0 += 56*c3p1*c5p6*ss[34];
  jac_3_0 += 56*c3p1*c7p6*ss[46];
  jac_3_0 += 168*c3p2*c5p5*ss[32];
  jac_3_0 += 168*c3p2*c7p5*ss[42];
  jac_3_0 += 280*c3p3*c5p4*ss[30];
  jac_3_0 += 280*c3p3*c7p4*ss[38];
  jac_3_0 += 280*c3p4*c5p3*ss[28];
  jac_3_0 += 280*c3p4*c7p3*ss[34];
  jac_3_0 += 168*c3p5*c5p2*ss[26];
  jac_3_0 += 168*c3p5*c7p2*ss[30];
  jac_3_0 += 56*c3p6*c5p1*ss[24];
  jac_3_0 += 56*c3p6*c7p1*ss[26];
  jac_3_0 += 56*c5p1*c7p6*ss[48];
  jac_3_0 += 168*c5p2*c7p5*ss[46];
  jac_3_0 += 280*c5p3*c7p4*ss[44];
  jac_3_0 += 280*c5p4*c7p3*ss[42];
  jac_3_0 += 168*c5p5*c7p2*ss[40];
  jac_3_0 += 56*c5p6*c7p1*ss[38];
  jac_3_0 += 8*c1p7*ss[8];
  jac_3_0 += 8*c3p7*ss[22];
  jac_3_0 += 8*c5p7*ss[36];
  jac_3_0 += 8*c7p7*ss[50];
  TACC jac_3_1 = convert<TACC>(1680*c1p1*c3p1*c5p1*c7p4*ss[40]);
  jac_3_1 += 3360*c1p1*c3p1*c5p2*c7p3*ss[38];
  jac_3_1 += 3360*c1p1*c3p1*c5p3*c7p2*ss[36];
  jac_3_1 += 1680*c1p1*c3p1*c5p4*c7p1*ss[34];
  jac_3_1 += 3360*c1p1*c3p2*c5p1*c7p3*ss[36];
  jac_3_1 += 5040*c1p1*c3p2*c5p2*c7p2*ss[34];
  jac_3_1 += 3360*c1p1*c3p2*c5p3*c7p1*ss[32];
  jac_3_1 += 3360*c1p1*c3p3*c5p1*c7p2*ss[32];
  jac_3_1 += 3360*c1p1*c3p3*c5p2*c7p1*ss[30];
  jac_3_1 += 1680*c1p1*c3p4*c5p1*c7p1*ss[28];
  jac_3_1 += 3360*c1p2*c3p1*c5p1*c7p3*ss[34];
  jac_3_1 += 5040*c1p2*c3p1*c5p2*c7p2*ss[32];
  jac_3_1 += 3360*c1p2*c3p1*c5p3*c7p1*ss[30];
  jac_3_1 += 5040*c1p2*c3p2*c5p1*c7p2*ss[30];
  jac_3_1 += 5040*c1p2*c3p2*c5p2*c7p1*ss[28];
  jac_3_1 += 3360*c1p2*c3p3*c5p1*c7p1*ss[26];
  jac_3_1 += 3360*c1p3*c3p1*c5p1*c7p2*ss[28];
  jac_3_1 += 3360*c1p3*c3p1*c5p2*c7p1*ss[26];
  jac_3_1 += 3360*c1p3*c3p2*c5p1*c7p1*ss[24];
  jac_3_1 += 1680*c1p4*c3p1*c5p1*c7p1*ss[22];
  jac_3_1 += 336*c1p1*c3p1*c5p5*ss[32];
  jac_3_1 += 336*c1p1*c3p1*c7p5*ss[42];
  jac_3_1 += 840*c1p1*c3p2*c5p4*ss[30];
  jac_3_1 += 840*c1p1*c3p2*c7p4*ss[38];
  jac_3_1 += 1120*c1p1*c3p3*c5p3*ss[28];
  jac_3_1 += 1120*c1p1*c3p3*c7p3*ss[34];
  jac_3_1 += 840*c1p1*c3p4*c5p2*ss[26];
  jac_3_1 += 840*c1p1*c3p4*c7p2*ss[30];
  jac_3_1 += 336*c1p1*c3p5*c5p1*ss[24];
  jac_3_1 += 336*c1p1*c3p5*c7p1*ss[26];
  jac_3_1 += 336*c1p1*c5p1*c7p5*ss[44];
  jac_3_1 += 840*c1p1*c5p2*c7p4*ss[42];
  jac_3_1 += 1120*c1p1*c5p3*c7p3*ss[40];
  jac_3_1 += 840*c1p1*c5p4*c7p2*ss[38];
  jac_3_1 += 336*c1p1*c5p5*c7p1*ss[36];
  jac_3_1 += 840*c1p2*c3p1*c5p4*ss[28];
  jac_3_1 += 840*c1p2*c3p1*c7p4*ss[36];
  jac_3_1 += 1680*c1p2*c3p2*c5p3*ss[26];
  jac_3_1 += 1680*c1p2*c3p2*c7p3*ss[32];
  jac_3_1 += 1680*c1p2*c3p3*c5p2*ss[24];
  jac_3_1 += 1680*c1p2*c3p3*c7p2*ss[28];
  jac_3_1 += 840*c1p2*c3p4*c5p1*ss[22];
  jac_3_1 += 840*c1p2*c3p4*c7p1*ss[24];
  jac_3_1 += 840*c1p2*c5p1*c7p4*ss[38];
  jac_3_1 += 1680*c1p2*c5p2*c7p3*ss[36];
  jac_3_1 += 1680*c1p2*c5p3*c7p2*ss[34];
  jac_3_1 += 840*c1p2*c5p4*c7p1*ss[32];
  jac_3_1 += 1120*c1p3*c3p1*c5p3*ss[24];
  jac_3_1 += 1120*c1p3*c3p1*c7p3*ss[30];
  jac_3_1 += 1680*c1p3*c3p2*c5p2*ss[22];
  jac_3_1 += 1680*c1p3*c3p2*c7p2*ss[26];
  jac_3_1 += 1120*c1p3*c3p3*c5p1*ss[20];
  jac_3_1 += 1120*c1p3*c3p3*c7p1*ss[22];
  jac_3_1 += 1120*c1p3*c5p1*c7p3*ss[32];
  jac_3_1 += 1680*c1p3*c5p2*c7p2*ss[30];
  jac_3_1 += 1120*c1p3*c5p3*c7p1*ss[28];
  jac_3_1 += 840*c1p4*c3p1*c5p2*ss[20];
  jac_3_1 += 840*c1p4*c3p1*c7p2*ss[24];
  jac_3_1 += 840*c1p4*c3p2*c5p1*ss[18];
  jac_3_1 += 840*c1p4*c3p2*c7p1*ss[20];
  jac_3_1 += 840*c1p4*c5p1*c7p2*ss[26];
  jac_3_1 += 840*c1p4*c5p2*c7p1*ss[24];
  jac_3_1 += 336*c1p5*c3p1*c5p1*ss[16];
  jac_3_1 += 336*c1p5*c3p1*c7p1*ss[18];
  jac_3_1 += 336*c1p5*c5p1*c7p1*ss[20];
  jac_3_1 += 336*c3p1*c5p1*c7p5*ss[46];
  jac_3_1 += 840*c3p1*c5p2*c7p4*ss[44];
  jac_3_1 += 1120*c3p1*c5p3*c7p3*ss[42];
  jac_3_1 += 840*c3p1*c5p4*c7p2*ss[40];
  jac_3_1 += 336*c3p1*c5p5*c7p1*ss[38];
  jac_3_1 += 840*c3p2*c5p1*c7p4*ss[42];
  jac_3_1 += 1680*c3p2*c5p2*c7p3*ss[40];
  jac_3_1 += 1680*c3p2*c5p3*c7p2*ss[38];
  jac_3_1 += 840*c3p2*c5p4*c7p1*ss[36];
  jac_3_1 += 1120*c3p3*c5p1*c7p3*ss[38];
  jac_3_1 += 1680*c3p3*c5p2*c7p2*ss[36];
  jac_3_1 += 1120*c3p3*c5p3*c7p1*ss[34];
  jac_3_1 += 840*c3p4*c5p1*c7p2*ss[34];
  jac_3_1 += 840*c3p4*c5p2*c7p1*ss[32];
  jac_3_1 += 336*c3p5*c5p1*c7p1*ss[30];
  jac_3_1 += 56*c1p1*c3p6*ss[22];
  jac_3_1 += 56*c1p1*c5p6*ss[34];
  jac_3_1 += 56*c1p1*c7p6*ss[46];
  jac_3_1 += 168*c1p2*c3p5*ss[20];
  jac_3_1 += 168*c1p2*c5p5*ss[30];
  jac_3_1 += 168*c1p2*c7p5*ss[40];
  jac_3_1 += 280*c1p3*c3p4*ss[18];
  jac_3_1 += 280*c1p3*c5p4*ss[26];
  jac_3_1 += 280*c1p3*c7p4*ss[34];
  jac_3_1 += 280*c1p4*c3p3*ss[16];
  jac_3_1 += 280*c1p4*c5p3*ss[22];
  jac_3_1 += 280*c1p4*c7p3*ss[28];
  jac_3_1 += 168*c1p5*c3p2*ss[14];
  jac_3_1 += 168*c1p5*c5p2*ss[18];
  jac_3_1 += 168*c1p5*c7p2*ss[22];
  jac_3_1 += 56*c1p6*c3p1*ss[12];
  jac_3_1 += 56*c1p6*c5p1*ss[14];
  jac_3_1 += 56*c1p6*c7p1*ss[16];
  jac_3_1 += 56*c3p1*c5p6*ss[36];
  jac_3_1 += 56*c3p1*c7p6*ss[48];
  jac_3_1 += 168*c3p2*c5p5*ss[34];
  jac_3_1 += 168*c3p2*c7p5*ss[44];
  jac_3_1 += 280*c3p3*c5p4*ss[32];
  jac_3_1 += 280*c3p3*c7p4*ss[40];
  jac_3_1 += 280*c3p4*c5p3*ss[30];
  jac_3_1 += 280*c3p4*c7p3*ss[36];
  jac_3_1 += 168*c3p5*c5p2*ss[28];
  jac_3_1 += 168*c3p5*c7p2*ss[32];
  jac_3_1 += 56*c3p6*c5p1*ss[26];
  jac_3_1 += 56*c3p6*c7p1*ss[28];
  jac_3_1 += 56*c5p1*c7p6*ss[50];
  jac_3_1 += 168*c5p2*c7p5*ss[48];
  jac_3_1 += 280*c5p3*c7p4*ss[46];
  jac_3_1 += 280*c5p4*c7p3*ss[44];
  jac_3_1 += 168*c5p5*c7p2*ss[42];
  jac_3_1 += 56*c5p6*c7p1*ss[40];
  jac_3_1 += 8*c1p7*ss[10];
  jac_3_1 += 8*c3p7*ss[24];
  jac_3_1 += 8*c5p7*ss[38];
  jac_3_1 += 8*c7p7*ss[52];
  TACC jac_3_2 = convert<TACC>(1680*c1p1*c3p1*c5p1*c7p4*ss[42]);
  jac_3_2 += 3360*c1p1*c3p1*c5p2*c7p3*ss[40];
  jac_3_2 += 3360*c1p1*c3p1*c5p3*c7p2*ss[38];
  jac_3_2 += 1680*c1p1*c3p1*c5p4*c7p1*ss[36];
  jac_3_2 += 3360*c1p1*c3p2*c5p1*c7p3*ss[38];
  jac_3_2 += 5040*c1p1*c3p2*c5p2*c7p2*ss[36];
  jac_3_2 += 3360*c1p1*c3p2*c5p3*c7p1*ss[34];
  jac_3_2 += 3360*c1p1*c3p3*c5p1*c7p2*ss[34];
  jac_3_2 += 3360*c1p1*c3p3*c5p2*c7p1*ss[32];
  jac_3_2 += 1680*c1p1*c3p4*c5p1*c7p1*ss[30];
  jac_3_2 += 3360*c1p2*c3p1*c5p1*c7p3*ss[36];
  jac_3_2 += 5040*c1p2*c3p1*c5p2*c7p2*ss[34];
  jac_3_2 += 3360*c1p2*c3p1*c5p3*c7p1*ss[32];
  jac_3_2 += 5040*c1p2*c3p2*c5p1*c7p2*ss[32];
  jac_3_2 += 5040*c1p2*c3p2*c5p2*c7p1*ss[30];
  jac_3_2 += 3360*c1p2*c3p3*c5p1*c7p1*ss[28];
  jac_3_2 += 3360*c1p3*c3p1*c5p1*c7p2*ss[30];
  jac_3_2 += 3360*c1p3*c3p1*c5p2*c7p1*ss[28];
  jac_3_2 += 3360*c1p3*c3p2*c5p1*c7p1*ss[26];
  jac_3_2 += 1680*c1p4*c3p1*c5p1*c7p1*ss[24];
  jac_3_2 += 336*c1p1*c3p1*c5p5*ss[34];
  jac_3_2 += 336*c1p1*c3p1*c7p5*ss[44];
  jac_3_2 += 840*c1p1*c3p2*c5p4*ss[32];
  jac_3_2 += 840*c1p1*c3p2*c7p4*ss[40];
  jac_3_2 += 1120*c1p1*c3p3*c5p3*ss[30];
  jac_3_2 += 1120*c1p1*c3p3*c7p3*ss[36];
  jac_3_2 += 840*c1p1*c3p4*c5p2*ss[28];
  jac_3_2 += 840*c1p1*c3p4*c7p2*ss[32];
  jac_3_2 += 336*c1p1*c3p5*c5p1*ss[26];
  jac_3_2 += 336*c1p1*c3p5*c7p1*ss[28];
  jac_3_2 += 336*c1p1*c5p1*c7p5*ss[46];
  jac_3_2 += 840*c1p1*c5p2*c7p4*ss[44];
  jac_3_2 += 1120*c1p1*c5p3*c7p3*ss[42];
  jac_3_2 += 840*c1p1*c5p4*c7p2*ss[40];
  jac_3_2 += 336*c1p1*c5p5*c7p1*ss[38];
  jac_3_2 += 840*c1p2*c3p1*c5p4*ss[30];
  jac_3_2 += 840*c1p2*c3p1*c7p4*ss[38];
  jac_3_2 += 1680*c1p2*c3p2*c5p3*ss[28];
  jac_3_2 += 1680*c1p2*c3p2*c7p3*ss[34];
  jac_3_2 += 1680*c1p2*c3p3*c5p2*ss[26];
  jac_3_2 += 1680*c1p2*c3p3*c7p2*ss[30];
  jac_3_2 += 840*c1p2*c3p4*c5p1*ss[24];
  jac_3_2 += 840*c1p2*c3p4*c7p1*ss[26];
  jac_3_2 += 840*c1p2*c5p1*c7p4*ss[40];
  jac_3_2 += 1680*c1p2*c5p2*c7p3*ss[38];
  jac_3_2 += 1680*c1p2*c5p3*c7p2*ss[36];
  jac_3_2 += 840*c1p2*c5p4*c7p1*ss[34];
  jac_3_2 += 1120*c1p3*c3p1*c5p3*ss[26];
  jac_3_2 += 1120*c1p3*c3p1*c7p3*ss[32];
  jac_3_2 += 1680*c1p3*c3p2*c5p2*ss[24];
  jac_3_2 += 1680*c1p3*c3p2*c7p2*ss[28];
  jac_3_2 += 1120*c1p3*c3p3*c5p1*ss[22];
  jac_3_2 += 1120*c1p3*c3p3*c7p1*ss[24];
  jac_3_2 += 1120*c1p3*c5p1*c7p3*ss[34];
  jac_3_2 += 1680*c1p3*c5p2*c7p2*ss[32];
  jac_3_2 += 1120*c1p3*c5p3*c7p1*ss[30];
  jac_3_2 += 840*c1p4*c3p1*c5p2*ss[22];
  jac_3_2 += 840*c1p4*c3p1*c7p2*ss[26];
  jac_3_2 += 840*c1p4*c3p2*c5p1*ss[20];
  jac_3_2 += 840*c1p4*c3p2*c7p1*ss[22];
  jac_3_2 += 840*c1p4*c5p1*c7p2*ss[28];
  jac_3_2 += 840*c1p4*c5p2*c7p1*ss[26];
  jac_3_2 += 336*c1p5*c3p1*c5p1*ss[18];
  jac_3_2 += 336*c1p5*c3p1*c7p1*ss[20];
  jac_3_2 += 336*c1p5*c5p1*c7p1*ss[22];
  jac_3_2 += 336*c3p1*c5p1*c7p5*ss[48];
  jac_3_2 += 840*c3p1*c5p2*c7p4*ss[46];
  jac_3_2 += 1120*c3p1*c5p3*c7p3*ss[44];
  jac_3_2 += 840*c3p1*c5p4*c7p2*ss[42];
  jac_3_2 += 336*c3p1*c5p5*c7p1*ss[40];
  jac_3_2 += 840*c3p2*c5p1*c7p4*ss[44];
  jac_3_2 += 1680*c3p2*c5p2*c7p3*ss[42];
  jac_3_2 += 1680*c3p2*c5p3*c7p2*ss[40];
  jac_3_2 += 840*c3p2*c5p4*c7p1*ss[38];
  jac_3_2 += 1120*c3p3*c5p1*c7p3*ss[40];
  jac_3_2 += 1680*c3p3*c5p2*c7p2*ss[38];
  jac_3_2 += 1120*c3p3*c5p3*c7p1*ss[36];
  jac_3_2 += 840*c3p4*c5p1*c7p2*ss[36];
  jac_3_2 += 840*c3p4*c5p2*c7p1*ss[34];
  jac_3_2 += 336*c3p5*c5p1*c7p1*ss[32];
  jac_3_2 += 56*c1p1*c3p6*ss[24];
  jac_3_2 += 56*c1p1*c5p6*ss[36];
  jac_3_2 += 56*c1p1*c7p6*ss[48];
  jac_3_2 += 168*c1p2*c3p5*ss[22];
  jac_3_2 += 168*c1p2*c5p5*ss[32];
  jac_3_2 += 168*c1p2*c7p5*ss[42];
  jac_3_2 += 280*c1p3*c3p4*ss[20];
  jac_3_2 += 280*c1p3*c5p4*ss[28];
  jac_3_2 += 280*c1p3*c7p4*ss[36];
  jac_3_2 += 280*c1p4*c3p3*ss[18];
  jac_3_2 += 280*c1p4*c5p3*ss[24];
  jac_3_2 += 280*c1p4*c7p3*ss[30];
  jac_3_2 += 168*c1p5*c3p2*ss[16];
  jac_3_2 += 168*c1p5*c5p2*ss[20];
  jac_3_2 += 168*c1p5*c7p2*ss[24];
  jac_3_2 += 56*c1p6*c3p1*ss[14];
  jac_3_2 += 56*c1p6*c5p1*ss[16];
  jac_3_2 += 56*c1p6*c7p1*ss[18];
  jac_3_2 += 56*c3p1*c5p6*ss[38];
  jac_3_2 += 56*c3p1*c7p6*ss[50];
  jac_3_2 += 168*c3p2*c5p5*ss[36];
  jac_3_2 += 168*c3p2*c7p5*ss[46];
  jac_3_2 += 280*c3p3*c5p4*ss[34];
  jac_3_2 += 280*c3p3*c7p4*ss[42];
  jac_3_2 += 280*c3p4*c5p3*ss[32];
  jac_3_2 += 280*c3p4*c7p3*ss[38];
  jac_3_2 += 168*c3p5*c5p2*ss[30];
  jac_3_2 += 168*c3p5*c7p2*ss[34];
  jac_3_2 += 56*c3p6*c5p1*ss[28];
  jac_3_2 += 56*c3p6*c7p1*ss[30];
  jac_3_2 += 56*c5p1*c7p6*ss[52];
  jac_3_2 += 168*c5p2*c7p5*ss[50];
  jac_3_2 += 280*c5p3*c7p4*ss[48];
  jac_3_2 += 280*c5p4*c7p3*ss[46];
  jac_3_2 += 168*c5p5*c7p2*ss[44];
  jac_3_2 += 56*c5p6*c7p1*ss[42];
  jac_3_2 += 8*c1p7*ss[12];
  jac_3_2 += 8*c3p7*ss[26];
  jac_3_2 += 8*c5p7*ss[40];
  jac_3_2 += 8*c7p7*ss[54];
  TACC jac_3_3 = convert<TACC>(1680*c1p1*c3p1*c5p1*c7p4*ss[44]);
  jac_3_3 += 3360*c1p1*c3p1*c5p2*c7p3*ss[42];
  jac_3_3 += 3360*c1p1*c3p1*c5p3*c7p2*ss[40];
  jac_3_3 += 1680*c1p1*c3p1*c5p4*c7p1*ss[38];
  jac_3_3 += 3360*c1p1*c3p2*c5p1*c7p3*ss[40];
  jac_3_3 += 5040*c1p1*c3p2*c5p2*c7p2*ss[38];
  jac_3_3 += 3360*c1p1*c3p2*c5p3*c7p1*ss[36];
  jac_3_3 += 3360*c1p1*c3p3*c5p1*c7p2*ss[36];
  jac_3_3 += 3360*c1p1*c3p3*c5p2*c7p1*ss[34];
  jac_3_3 += 1680*c1p1*c3p4*c5p1*c7p1*ss[32];
  jac_3_3 += 3360*c1p2*c3p1*c5p1*c7p3*ss[38];
  jac_3_3 += 5040*c1p2*c3p1*c5p2*c7p2*ss[36];
  jac_3_3 += 3360*c1p2*c3p1*c5p3*c7p1*ss[34];
  jac_3_3 += 5040*c1p2*c3p2*c5p1*c7p2*ss[34];
  jac_3_3 += 5040*c1p2*c3p2*c5p2*c7p1*ss[32];
  jac_3_3 += 3360*c1p2*c3p3*c5p1*c7p1*ss[30];
  jac_3_3 += 3360*c1p3*c3p1*c5p1*c7p2*ss[32];
  jac_3_3 += 3360*c1p3*c3p1*c5p2*c7p1*ss[30];
  jac_3_3 += 3360*c1p3*c3p2*c5p1*c7p1*ss[28];
  jac_3_3 += 1680*c1p4*c3p1*c5p1*c7p1*ss[26];
  jac_3_3 += 336*c1p1*c3p1*c5p5*ss[36];
  jac_3_3 += 336*c1p1*c3p1*c7p5*ss[46];
  jac_3_3 += 840*c1p1*c3p2*c5p4*ss[34];
  jac_3_3 += 840*c1p1*c3p2*c7p4*ss[42];
  jac_3_3 += 1120*c1p1*c3p3*c5p3*ss[32];
  jac_3_3 += 1120*c1p1*c3p3*c7p3*ss[38];
  jac_3_3 += 840*c1p1*c3p4*c5p2*ss[30];
  jac_3_3 += 840*c1p1*c3p4*c7p2*ss[34];
  jac_3_3 += 336*c1p1*c3p5*c5p1*ss[28];
  jac_3_3 += 336*c1p1*c3p5*c7p1*ss[30];
  jac_3_3 += 336*c1p1*c5p1*c7p5*ss[48];
  jac_3_3 += 840*c1p1*c5p2*c7p4*ss[46];
  jac_3_3 += 1120*c1p1*c5p3*c7p3*ss[44];
  jac_3_3 += 840*c1p1*c5p4*c7p2*ss[42];
  jac_3_3 += 336*c1p1*c5p5*c7p1*ss[40];
  jac_3_3 += 840*c1p2*c3p1*c5p4*ss[32];
  jac_3_3 += 840*c1p2*c3p1*c7p4*ss[40];
  jac_3_3 += 1680*c1p2*c3p2*c5p3*ss[30];
  jac_3_3 += 1680*c1p2*c3p2*c7p3*ss[36];
  jac_3_3 += 1680*c1p2*c3p3*c5p2*ss[28];
  jac_3_3 += 1680*c1p2*c3p3*c7p2*ss[32];
  jac_3_3 += 840*c1p2*c3p4*c5p1*ss[26];
  jac_3_3 += 840*c1p2*c3p4*c7p1*ss[28];
  jac_3_3 += 840*c1p2*c5p1*c7p4*ss[42];
  jac_3_3 += 1680*c1p2*c5p2*c7p3*ss[40];
  jac_3_3 += 1680*c1p2*c5p3*c7p2*ss[38];
  jac_3_3 += 840*c1p2*c5p4*c7p1*ss[36];
  jac_3_3 += 1120*c1p3*c3p1*c5p3*ss[28];
  jac_3_3 += 1120*c1p3*c3p1*c7p3*ss[34];
  jac_3_3 += 1680*c1p3*c3p2*c5p2*ss[26];
  jac_3_3 += 1680*c1p3*c3p2*c7p2*ss[30];
  jac_3_3 += 1120*c1p3*c3p3*c5p1*ss[24];
  jac_3_3 += 1120*c1p3*c3p3*c7p1*ss[26];
  jac_3_3 += 1120*c1p3*c5p1*c7p3*ss[36];
  jac_3_3 += 1680*c1p3*c5p2*c7p2*ss[34];
  jac_3_3 += 1120*c1p3*c5p3*c7p1*ss[32];
  jac_3_3 += 840*c1p4*c3p1*c5p2*ss[24];
  jac_3_3 += 840*c1p4*c3p1*c7p2*ss[28];
  jac_3_3 += 840*c1p4*c3p2*c5p1*ss[22];
  jac_3_3 += 840*c1p4*c3p2*c7p1*ss[24];
  jac_3_3 += 840*c1p4*c5p1*c7p2*ss[30];
  jac_3_3 += 840*c1p4*c5p2*c7p1*ss[28];
  jac_3_3 += 336*c1p5*c3p1*c5p1*ss[20];
  jac_3_3 += 336*c1p5*c3p1*c7p1*ss[22];
  jac_3_3 += 336*c1p5*c5p1*c7p1*ss[24];
  jac_3_3 += 336*c3p1*c5p1*c7p5*ss[50];
  jac_3_3 += 840*c3p1*c5p2*c7p4*ss[48];
  jac_3_3 += 1120*c3p1*c5p3*c7p3*ss[46];
  jac_3_3 += 840*c3p1*c5p4*c7p2*ss[44];
  jac_3_3 += 336*c3p1*c5p5*c7p1*ss[42];
  jac_3_3 += 840*c3p2*c5p1*c7p4*ss[46];
  jac_3_3 += 1680*c3p2*c5p2*c7p3*ss[44];
  jac_3_3 += 1680*c3p2*c5p3*c7p2*ss[42];
  jac_3_3 += 840*c3p2*c5p4*c7p1*ss[40];
  jac_3_3 += 1120*c3p3*c5p1*c7p3*ss[42];
  jac_3_3 += 1680*c3p3*c5p2*c7p2*ss[40];
  jac_3_3 += 1120*c3p3*c5p3*c7p1*ss[38];
  jac_3_3 += 840*c3p4*c5p1*c7p2*ss[38];
  jac_3_3 += 840*c3p4*c5p2*c7p1*ss[36];
  jac_3_3 += 336*c3p5*c5p1*c7p1*ss[34];
  jac_3_3 += 56*c1p1*c3p6*ss[26];
  jac_3_3 += 56*c1p1*c5p6*ss[38];
  jac_3_3 += 56*c1p1*c7p6*ss[50];
  jac_3_3 += 168*c1p2*c3p5*ss[24];
  jac_3_3 += 168*c1p2*c5p5*ss[34];
  jac_3_3 += 168*c1p2*c7p5*ss[44];
  jac_3_3 += 280*c1p3*c3p4*ss[22];
  jac_3_3 += 280*c1p3*c5p4*ss[30];
  jac_3_3 += 280*c1p3*c7p4*ss[38];
  jac_3_3 += 280*c1p4*c3p3*ss[20];
  jac_3_3 += 280*c1p4*c5p3*ss[26];
  jac_3_3 += 280*c1p4*c7p3*ss[32];
  jac_3_3 += 168*c1p5*c3p2*ss[18];
  jac_3_3 += 168*c1p5*c5p2*ss[22];
  jac_3_3 += 168*c1p5*c7p2*ss[26];
  jac_3_3 += 56*c1p6*c3p1*ss[16];
  jac_3_3 += 56*c1p6*c5p1*ss[18];
  jac_3_3 += 56*c1p6*c7p1*ss[20];
  jac_3_3 += 56*c3p1*c5p6*ss[40];
  jac_3_3 += 56*c3p1*c7p6*ss[52];
  jac_3_3 += 168*c3p2*c5p5*ss[38];
  jac_3_3 += 168*c3p2*c7p5*ss[48];
  jac_3_3 += 280*c3p3*c5p4*ss[36];
  jac_3_3 += 280*c3p3*c7p4*ss[44];
  jac_3_3 += 280*c3p4*c5p3*ss[34];
  jac_3_3 += 280*c3p4*c7p3*ss[40];
  jac_3_3 += 168*c3p5*c5p2*ss[32];
  jac_3_3 += 168*c3p5*c7p2*ss[36];
  jac_3_3 += 56*c3p6*c5p1*ss[30];
  jac_3_3 += 56*c3p6*c7p1*ss[32];
  jac_3_3 += 56*c5p1*c7p6*ss[54];
  jac_3_3 += 168*c5p2*c7p5*ss[52];
  jac_3_3 += 280*c5p3*c7p4*ss[50];
  jac_3_3 += 280*c5p4*c7p3*ss[48];
  jac_3_3 += 168*c5p5*c7p2*ss[46];
  jac_3_3 += 56*c5p6*c7p1*ss[44];
  jac_3_3 += 8*c1p7*ss[14];
  jac_3_3 += 8*c3p7*ss[28];
  jac_3_3 += 8*c5p7*ss[42];
  jac_3_3 += 8*c7p7*ss[56];
  roots[0]=convert<T>(root_0);
  roots[1]=convert<T>(root_1);
  roots[2]=convert<T>(root_2);
  roots[3]=convert<T>(root_3);
  jacobian[0] = convert<T>(jac_0_0);
  jacobian[1] = convert<T>(jac_0_1);
  jacobian[2] = convert<T>(jac_0_2);
  jacobian[3] = convert<T>(jac_0_3);
  jacobian[4] = convert<T>(jac_1_0);
  jacobian[5] = convert<T>(jac_1_1);
  jacobian[6] = convert<T>(jac_1_2);
  jacobian[7] = convert<T>(jac_1_3);
  jacobian[8] = convert<T>(jac_2_0);
  jacobian[9] = convert<T>(jac_2_1);
  jacobian[10] = convert<T>(jac_2_2);
  jacobian[11] = convert<T>(jac_2_3);
  jacobian[12] = convert<T>(jac_3_0);
  jacobian[13] = convert<T>(jac_3_1);
  jacobian[14] = convert<T>(jac_3_2);
  jacobian[15] = convert<T>(jac_3_3);
}
}; // SymmetricHepticSystem

}; // systems
}; // random
}; // flopoco

#endif
